"""
Planetary Model
"""

# Django
from django.db import models
from django.utils import timezone
from django.utils.dateparse import parse_datetime
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EvePlanet, EveType

# AA Ledger
from ledger import __title__
from ledger.constants import EXTRACTOR_CONTROL_UNIT, P0_PRODUCTS, SPACEPORTS
from ledger.managers.character_planetary_manager import (
    PlanetaryDetailsManager,
    PlanetaryManager,
)
from ledger.models.characteraudit import CharacterAudit, CharacterUpdateStatus

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharacterPlanet(models.Model):
    objects = PlanetaryManager()

    id = models.AutoField(primary_key=True)

    planet_name = models.CharField(max_length=100, null=True, default=None)

    planet = models.ForeignKey(
        EvePlanet, on_delete=models.CASCADE, related_name="ledger_planet"
    )

    character = models.ForeignKey(
        CharacterAudit, on_delete=models.CASCADE, related_name="ledger_character_planet"
    )

    upgrade_level = models.IntegerField(
        default=0, help_text=_("Upgrade level of the planet")
    )

    num_pins = models.IntegerField(
        default=0, help_text=_("Number of pins on the planet")
    )

    class Meta:
        default_permissions = ()
        indexes = [
            models.Index(fields=["character"]),
            models.Index(fields=["planet"]),
        ]

    def __str__(self):
        return f"Planet Data: {self.character.eve_character.character_name} - {self.planet.name}"

    @property
    def last_update(self) -> timezone.datetime:
        """Return the last update time of the planet."""
        try:
            last_update = CharacterUpdateStatus.objects.get(
                character=self.character,
                section=CharacterAudit.UpdateSection.PLANETS,
            ).last_update_at
        except CharacterUpdateStatus.DoesNotExist:
            last_update = None
        return last_update

    @classmethod
    def get_esi_scopes(cls) -> list[str]:
        """Return list of required ESI scopes to fetch."""
        return [
            "esi-planets.manage_planets.v1",
        ]


class CharacterPlanetDetails(models.Model):
    """Model to store the details of a planet"""

    objects = PlanetaryDetailsManager()

    id = models.AutoField(primary_key=True)

    planet = models.ForeignKey(
        CharacterPlanet,
        on_delete=models.CASCADE,
        related_name="ledger_planet_details",
    )

    character = models.ForeignKey(
        CharacterAudit,
        on_delete=models.CASCADE,
        related_name="ledger_character_planet_details",
    )

    links = models.JSONField(null=True, default=None, blank=True)
    pins = models.JSONField(null=True, default=None, blank=True)
    routes = models.JSONField(null=True, default=None, blank=True)
    facilitys = models.JSONField(null=True, default=None, blank=True)

    last_alert = models.DateTimeField(null=True, default=None, blank=True)

    notification = models.BooleanField(default=False)
    notification_sent = models.BooleanField(default=False)

    class Meta:
        default_permissions = ()
        indexes = [
            models.Index(fields=["planet"]),
        ]

    def __str__(self):
        return f"Planet Details Data: {self.planet.character.eve_character.character_name} - {self.planet.planet.name}"

    def count_extractors(self):
        return len(
            [pin for pin in self.pins if pin.get("type_id") in EXTRACTOR_CONTROL_UNIT]
        )

    def get_planet_install_date(self):
        install_times = [
            pin.get("install_time")
            for pin in self.pins
            if pin.get("install_time") and pin["install_time"] != "0"
        ]
        if install_times:
            install = timezone.datetime.fromisoformat(
                min(install_times).replace("Z", "+00:00")
            )
            return install
        return None

    def get_planet_expiry_date(self):
        if self.pins is None:
            return None

        expiry_times = [
            pin.get("expiry_time")
            for pin in self.pins
            if pin.get("expiry_time") and pin["expiry_time"] != "0"
        ]
        if expiry_times:
            alert = timezone.datetime.fromisoformat(
                min(expiry_times).replace("Z", "+00:00")
            )
            return alert
        return None

    def get_types(self) -> list:
        """Get the product types of the routes on the planet"""
        types = []
        for pin in self.routes:
            if pin.get("content_type_id") not in types:
                types.append(pin.get("content_type_id"))
        return types

    def allocate_products(self) -> dict:
        """Get the product types on the planet"""
        product_types = {}
        for c_type_id in self.routes:
            type_id = c_type_id.get("content_type_id")
            if type_id not in P0_PRODUCTS and type_id not in product_types:
                type_data, _ = EveType.objects.get_or_create_esi(id=type_id)
                product_types[type_id] = {
                    "id": type_id,
                    "name": type_data.name,
                    "category": type_data.eve_group.name,
                }
        return product_types

    def allocate_extracts(self) -> dict:
        """Get the extractor raw product types on the planet"""
        product_types = {}
        for c_type_id in self.routes:
            type_id = c_type_id.get("content_type_id")
            if type_id in P0_PRODUCTS and type_id not in product_types:
                type_data, _ = EveType.objects.get_or_create_esi(id=type_id)
                product_types[type_id] = {
                    "id": type_id,
                    "name": type_data.name,
                    "category": type_data.eve_group.name,
                }
        return product_types

    def allocate_overall_progress(self) -> dict:
        extractors = self.get_extractors_info()

        if not extractors:
            return 0

        total_install_time = 0
        total_expiry_time = 0
        current_time = timezone.now().timestamp() * 1000  # Current time in milliseconds
        extractor_count = len(extractors)

        for extractor in extractors.values():
            if not extractor.get("install_time") or not extractor.get("expiry_time"):
                continue
            install_time = (
                timezone.datetime.fromisoformat(extractor["install_time"]).timestamp()
                * 1000
            )
            expiry_time = (
                timezone.datetime.fromisoformat(extractor["expiry_time"]).timestamp()
                * 1000
            )

            total_install_time += install_time
            total_expiry_time += expiry_time

        average_install_time = total_install_time / extractor_count
        average_expiry_time = total_expiry_time / extractor_count

        total_duration = average_expiry_time - average_install_time
        elapsed_duration = current_time - average_install_time
        progress_percentage = min(
            max((elapsed_duration / total_duration) * 100, 0), 100
        )

        return progress_percentage

    def get_extractors_info(self) -> dict:
        extractors = {}
        current_time = timezone.now().timestamp()

        for pin in self.pins:
            extractor_details = pin.get("extractor_details")
            if extractor_details and "cycle_time" in extractor_details:
                type_id = extractor_details.get("product_type_id")
                type_data, _ = EveType.objects.get_or_create_esi(id=type_id)

                install_time_str = pin.get("install_time")
                expiry_time_str = pin.get("expiry_time")

                install_time = parse_datetime(install_time_str).timestamp()
                expiry_time = parse_datetime(expiry_time_str).timestamp()

                # Calculate progress percentage (0-100)
                if current_time >= expiry_time:
                    progress_percentage = 100.0
                else:
                    progress_percentage = round(
                        ((current_time - install_time) / (expiry_time - install_time))
                        * 100,
                        2,
                    )

                extractors[pin.get("pin_id")] = {
                    "install_time": install_time_str,
                    "expiry_time": expiry_time_str,
                    "item_id": type_id,
                    "item_name": type_data.name,
                    "progress_percentage": progress_percentage,
                }
        return extractors

    def get_storage_info(self) -> dict:
        storage = {}
        for pin in self.pins:
            type_id = pin.get("type_id")
            if type_id in SPACEPORTS:
                type_data, _ = EveType.objects.get_or_create_esi(id=type_id)
                contents_info = []
                for content in pin.get("contents", []):
                    content_type_id = content.get("type_id")
                    content_type_data, _ = EveType.objects.get_or_create_esi(
                        id=content_type_id
                    )
                    contents_info.append(
                        {
                            "amount": content.get("amount"),
                            "type_id": content_type_id,
                            "product_name": content_type_data.name,
                        }
                    )
                storage[pin.get("pin_id")] = {
                    "facility_id": type_id,
                    "facility_name": type_data.name,
                    "contents": contents_info,
                }
        return storage

    @property
    def is_expired(self):
        expiry_date = self.get_planet_expiry_date()
        if expiry_date is None:
            return False
        return expiry_date < timezone.now()

    @property
    def last_update(self) -> timezone.datetime:
        """Return the last update time of the planet details."""
        try:
            last_update = CharacterUpdateStatus.objects.get(
                character=self.character,
                section=CharacterAudit.UpdateSection.PLANETS_DETAILS,
            ).last_update_at
        except CharacterUpdateStatus.DoesNotExist:
            last_update = None
        return last_update



# Standard Library
from unittest.mock import Mock, patch

# Django
from django.contrib.admin.sites import AdminSite
from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, override_settings
from django.utils import timezone
from django.utils.safestring import mark_safe

# Alliance Auth
from allianceauth.eveonline.evelinks import eveimageserver

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase

# AA Ledger
from ledger.admin import (
    CharacterAuditAdmin,
    CharacterUpdateStatusAdminInline,
    CorporationAuditAdmin,
    CorporationUpdateStatusAdminInline,
)
from ledger.models.characteraudit import CharacterAudit, CharacterUpdateStatus
from ledger.models.corporationaudit import CorporationAudit, CorporationUpdateStatus
from ledger.tests.testdata.generate_characteraudit import (
    add_characteraudit_character_to_user,
    create_update_status,
)
from ledger.tests.testdata.generate_corporationaudit import (
    add_corporationaudit_corporation_to_user,
    create_corporation_update_status,
    create_user_from_evecharacter,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth

ADMIN_PATH = "ledger.admin"


class MockRequest:
    pass


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestCorporationAuditAdmin(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()

        cls.factory = RequestFactory()
        cls.site = AdminSite()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            ["ledger.basic_access"],
        )
        cls.corporation_audit_admin = CorporationAuditAdmin(CorporationAudit, cls.site)
        cls.corporation_audit = add_corporationaudit_corporation_to_user(
            cls.user,
            cls.character_ownership.character.character_id,
        )
        cls.user.is_superuser = True

        cls.update_status = create_corporation_update_status(
            cls.corporation_audit,
            section="wallet_journal",
            last_update_at=timezone.now() - timezone.timedelta(minutes=5),
            last_update_finished_at=timezone.now() - timezone.timedelta(minutes=3),
            last_run_at=timezone.now() - timezone.timedelta(minutes=4),
            last_run_finished_at=timezone.now() - timezone.timedelta(minutes=2),
        )

    def test_get_queryset(self):
        qs = self.corporation_audit_admin.get_queryset(CorporationAudit.objects.all())
        self.assertIsNotNone(qs)

    def test_run_update_duration(self):
        inline = CorporationUpdateStatusAdminInline(CorporationUpdateStatus, self.site)

        run_duration = inline._run_duration(self.update_status)
        update_duration = inline._update_duration(self.update_status).replace(
            "\xa0", " "
        )
        self.assertIn("minute", run_duration)
        self.assertIn("2 minutes", update_duration)
        self.assertEqual(inline._calc_duration(None, None), "-")

    @patch(ADMIN_PATH + ".update_corporation.delay")
    def test_force_update(self, mock_update_character_delay):
        request = self.factory.get("/")
        request.user = self.user

        # Add session middleware to process the request
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        message_middleware = MessageMiddleware(Mock())
        message_middleware.process_request(request)

        queryset = CorporationAudit.objects.filter(pk=self.corporation_audit.pk)
        self.corporation_audit_admin.force_update(request, queryset)
        mock_update_character_delay.assert_called_once()

    def test_entity_pic(self):
        self.client.force_login(self.user)
        request = self.factory.get("/")
        request.user = self.user
        expected_html = '<img src="{}" class="img-circle">'.format(
            eveimageserver._eve_entity_image_url(
                "corporation", self.corporation_audit.corporation.corporation_id, 32
            )
        )
        self.assertEqual(
            self.corporation_audit_admin._entity_pic(self.corporation_audit),
            expected_html,
        )

    def test_corporation_corporation_id(self):
        self.client.force_login(self.user)
        request = self.factory.get("/")
        request.user = self.user
        self.assertEqual(
            self.corporation_audit_admin._corporation__corporation_id(
                self.corporation_audit
            ),
            2001,
        )

    def test_has_add_permission(self):
        self.client.force_login(self.user)
        request = self.factory.get("/")
        request.user = self.user
        self.assertFalse(self.corporation_audit_admin.has_add_permission(request))

    def test_has_change_permission(self):
        self.client.force_login(self.user)
        request = self.factory.get("/")
        request.user = self.user
        request = MockRequest()
        self.assertFalse(self.corporation_audit_admin.has_change_permission(request))
        self.assertFalse(
            self.corporation_audit_admin.has_change_permission(
                request, obj=self.corporation_audit
            )
        )


class TestCharacterAuditAdmin(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        cls.factory = RequestFactory()
        cls.site = AdminSite()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            ["ledger.basic_access"],
        )
        cls.character_audit_admin = CharacterAuditAdmin(CharacterAudit, cls.site)
        cls.character_audit = add_characteraudit_character_to_user(
            cls.user,
            cls.character_ownership.character.character_id,
        )
        cls.user.is_superuser = True
        cls.update_status = create_update_status(
            cls.character_audit,
            section="wallet_journal",
            last_update_at=timezone.now() - timezone.timedelta(minutes=5),
            last_update_finished_at=timezone.now() - timezone.timedelta(minutes=3),
            last_run_at=timezone.now() - timezone.timedelta(minutes=4),
            last_run_finished_at=timezone.now() - timezone.timedelta(minutes=2),
        )

    def test_get_queryset(self):
        qs = self.character_audit_admin.get_queryset(CharacterAudit.objects.all())
        self.assertIsNotNone(qs)

    @patch(ADMIN_PATH + ".update_character.delay")
    def test_force_update(self, mock_update_character_delay):
        queryset = CharacterAudit.objects.filter(pk=self.character_audit.pk)
        request = self.factory.get("/")

        # Add session middleware to process the request
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        message_middleware = MessageMiddleware(Mock())
        message_middleware.process_request(request)

        self.character_audit_admin.force_update(request, queryset)
        mock_update_character_delay.assert_called_once()

    def test_run_update_duration(self):
        inline = CharacterUpdateStatusAdminInline(CharacterUpdateStatus, self.site)

        run_duration = inline._run_duration(self.update_status)
        update_duration = inline._update_duration(self.update_status).replace(
            "\xa0", " "
        )
        self.assertIn("minute", run_duration)
        self.assertIn("2 minutes", update_duration)
        self.assertEqual(inline._calc_duration(None, None), "-")

    def test_entity_pic(self):
        self.client.force_login(self.user)
        expected_html = '<img src="{}" class="img-circle">'.format(
            eveimageserver._eve_entity_image_url(
                "character", self.character_audit.eve_character.character_id, 32
            )
        )
        self.assertEqual(
            self.character_audit_admin._entity_pic(self.character_audit), expected_html
        )

    def test_character_character_name(self):
        self.assertEqual(
            self.character_audit_admin._eve_character__character_name(
                self.character_audit
            ),
            self.character_audit.eve_character.character_name,
        )

    def test_has_add_permission(self):
        request = self.factory.get("/")
        self.assertFalse(self.character_audit_admin.has_add_permission(request))

    def test_has_change_permission(self):
        request = self.factory.get("/")
        self.assertFalse(self.character_audit_admin.has_change_permission(request))
        self.assertFalse(
            self.character_audit_admin.has_change_permission(
                request, obj=self.character_audit
            )
        )



# AA Ledger
from ledger.models.characteraudit import CharacterAudit, CharacterMiningLedger


def create_miningledger(character: CharacterAudit, **kwargs) -> CharacterMiningLedger:
    """Create a CorporationWalletDivision"""
    params = {
        "character": character,
    }
    params.update(kwargs)
    division = CharacterMiningLedger(**params)
    division.save()
    return division



# Django
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import QuerySet

# Alliance Auth
from allianceauth.eveonline.models import EveAllianceInfo
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__, models

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def get_character_or_none(
    request, character_id
) -> tuple[bool, models.CharacterAudit | None]:
    """Get Character and check permissions"""
    perms = True

    try:
        character = models.CharacterAudit.objects.get(
            eve_character__character_id=character_id
        )
    except ObjectDoesNotExist:
        return False, None
    except ValueError:
        return None, None

    # check access
    visible = models.CharacterAudit.objects.visible_eve_characters(request.user)
    if character.eve_character not in visible:
        perms = False
    return perms, character


def get_corporation(
    request, corporation_id
) -> tuple[bool | None, models.CorporationAudit | None]:
    """Return Corporation and check permissions"""
    perms = True

    try:
        main_corp = models.CorporationAudit.objects.get(
            corporation__corporation_id=corporation_id
        )
    except ObjectDoesNotExist:
        return None, None

    # Check access
    visible = models.CorporationAudit.objects.visible_to(request.user)
    if main_corp not in visible:
        perms = False
    return perms, main_corp


def get_manage_corporation(
    request, corporation_id
) -> tuple[bool | None, models.CorporationAudit | None]:
    """Returns a tuple of the permissions and the corporation object if manageable"""
    perms = True

    try:
        main_corp = models.CorporationAudit.objects.get(
            corporation__corporation_id=corporation_id
        )
    except ObjectDoesNotExist:
        return None, None

    # Check access
    visible = models.CorporationAudit.objects.manage_to(request.user)
    if main_corp not in visible:
        perms = False
    return perms, main_corp


def get_alliance(request, alliance_id) -> tuple[bool | None, EveAllianceInfo | None]:
    """Get Alliance and check permissions for each corporation"""
    perms = True

    corporations = models.CorporationAudit.objects.filter(
        corporation__alliance__alliance_id=alliance_id
    )

    if not corporations.exists():
        return None, None

    # Check access
    visible = models.CorporationAudit.objects.visible_to(request.user)

    # Check if there is an intersection between main_corp and visible
    common_corps = corporations.intersection(visible)
    if not common_corps.exists():
        perms = False

    ally = EveAllianceInfo.objects.get(alliance_id=alliance_id)
    return perms, ally


def get_all_corporations_from_alliance(
    request, alliance_id
) -> tuple[bool | None, list[models.CorporationAudit] | None]:
    """Get Alliance and check permissions for each corporation"""
    perms = True

    corporations = models.CorporationAudit.objects.filter(
        corporation__alliance__alliance_id=alliance_id
    )

    if not corporations.exists():
        return None, None

    # Check access
    visible = models.CorporationAudit.objects.visible_to(request.user)

    # Check if there is an intersection between main_corp and visible
    common_corps = corporations.intersection(visible)
    if not common_corps.exists():
        perms = False
    return perms, corporations


def get_alts_queryset(
    character: models.CharacterAudit,
) -> QuerySet[list[models.CharacterAudit]]:
    """Get all alts for a main character."""
    try:
        linked_characters = (
            character.eve_character.character_ownership.user.character_ownerships.all()
        )

        linked_characters = linked_characters.values_list("character_id", flat=True)
        return models.CharacterAudit.objects.filter(
            eve_character__id__in=linked_characters
        )
    except ObjectDoesNotExist:
        return models.CharacterAudit.objects.filter(
            eve_character__pk=character.eve_character.pk
        )



# Standard Library
import calendar

# Django
from django.template.defaultfilters import register
from django.utils import timezone
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__, __version__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@register.filter
def range_filter(value):
    return range(1, value + 1)


@register.filter
def month_name(month):
    """Returns the month name for a given month number (1-12), übersetzbar."""
    months = [
        _("January"),
        _("February"),
        _("March"),
        _("April"),
        _("May"),
        _("June"),
        _("July"),
        _("August"),
        _("September"),
        _("October"),
        _("November"),
        _("December"),
    ]
    try:
        if 1 <= month <= 12:
            return months[month - 1]
    except (IndexError, ValueError, TypeError):
        pass
    return ""


@register.simple_tag
def month_days(date_info):
    """Returns the number of days in a given month, fallback to current if missing."""
    # Versuche, year und month aus dem Argument zu holen (dict oder Objekt)
    year = (
        getattr(date_info, "year", None) or date_info.get("year")
        if hasattr(date_info, "get")
        else None
    )
    month = (
        getattr(date_info, "month", None) or date_info.get("month")
        if hasattr(date_info, "get")
        else None
    )

    # Fallback auf aktuelles Datum, falls None
    now = timezone.now()
    year = year or now.year
    month = month or now.month

    try:
        return list(range(1, calendar.monthrange(int(year), int(month))[1] + 1))
    except Exception:  # pylint: disable=broad-except
        return []


@register.filter
def get_item(dictionary, key):
    """Returns the value for a given key in a dictionary, or empty dict if not found."""
    if isinstance(dictionary, dict):
        return dictionary.get(key, {})
    return {}



# Generated by Django 4.2.11 on 2024-06-07 13:41

# Django
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0001_initial"),
    ]

    operations = [
        migrations.DeleteModel(
            name="CorpSteuer",
        ),
        migrations.AlterModelOptions(
            name="characteraudit",
            options={
                "default_permissions": (),
                "permissions": (
                    ("char_audit_admin_access", "Has access to all characters"),
                    ("char_audit_manager", "Has access to all characters for own Corp"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="corporationaudit",
            options={
                "default_permissions": (),
                "permissions": (
                    ("corp_audit_admin_access", "Has access to all Corporations"),
                    ("corp_audit_manager", "Has access to own Corporation"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    ("basic_access", "Can access this app, Ledger."),
                    ("admin_access", "Has access to all Administration tools"),
                ),
            },
        ),
    ]



"""Admin models"""

# Django
from django.contrib import admin, messages
from django.contrib.humanize.templatetags.humanize import naturaltime
from django.db.models import Max, Q
from django.utils import timezone
from django.utils.html import format_html
from django.utils.timesince import timesince
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks import eveimageserver

# AA Ledger
from ledger.models.characteraudit import CharacterAudit, CharacterUpdateStatus
from ledger.models.corporationaudit import CorporationAudit, CorporationUpdateStatus
from ledger.tasks import update_character, update_corporation


class CorporationUpdateStatusAdminInline(admin.TabularInline):
    model = CorporationUpdateStatus
    fields = (
        "section",
        "_is_enabled",
        "_is_success",
        "_is_token_ok",
        "error_message",
        "last_run_finished_at",
        "_run_duration",
        "last_update_finished_at",
        "_update_duration",
    )
    readonly_fields = (
        "_is_enabled",
        "_is_success",
        "_is_token_ok",
        "_run_duration",
        "_update_duration",
    )
    ordering = ["section"]

    # pylint: disable=unused-argument
    def has_add_permission(self, request, obj=None):
        return False

    # pylint: disable=unused-argument
    def has_change_permission(self, request, obj=None):
        return False

    # pylint: disable=unused-argument
    def has_delete_permission(self, request, obj=None):
        return False

    @admin.display(boolean=True)
    def _is_enabled(self, obj: CorporationUpdateStatus) -> bool:
        return obj.is_enabled

    @admin.display(boolean=True)
    def _is_success(self, obj: CorporationUpdateStatus) -> bool:
        if not obj.is_enabled:
            return None
        return obj.is_success

    @admin.display(boolean=True)
    def _is_token_ok(self, obj: CorporationUpdateStatus) -> bool:
        return not obj.has_token_error

    @admin.display
    def _run_duration(self, obj: CorporationUpdateStatus) -> float:
        return self._calc_duration(obj.last_run_at, obj.last_run_finished_at)

    @admin.display
    def _update_duration(self, obj: CorporationUpdateStatus) -> float:
        return self._calc_duration(obj.last_update_at, obj.last_update_finished_at)

    @staticmethod
    def _calc_duration(
        started_at: timezone.datetime, finished_at: timezone.datetime
    ) -> timezone.timedelta:
        if not started_at or not finished_at:
            return "-"

        return timesince(started_at, finished_at)


@admin.register(CorporationAudit)
class CorporationAuditAdmin(admin.ModelAdmin):
    list_display = (
        "_entity_pic",
        "_corporation__corporation_id",
        "_last_update_at",
    )

    list_display_links = (
        "_entity_pic",
        "_corporation__corporation_id",
    )

    list_select_related = ("corporation",)

    ordering = ["corporation__corporation_name"]

    search_fields = ["corporation__corporation_name", "corporation__corporation_id"]

    actions = [
        "delete_objects",
        "force_update",
    ]

    inlines = (CorporationUpdateStatusAdminInline,)

    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)
        return qs.prefetch_related("ledger_corporation_update_status").annotate(
            last_update_at=Max(
                "ledger_corporation_update_status__last_run_finished_at",
                filter=~Q(ledger_corporation_update_status__section="wallet_journal"),
            )
        )

    @admin.display(description="")
    def _entity_pic(self, obj: CorporationAudit):
        eve_id = obj.corporation.corporation_id
        return format_html(
            '<img src="{}" class="img-circle">',
            eveimageserver._eve_entity_image_url("corporation", eve_id, 32),
        )

    @admin.display(description="Corporation ID", ordering="corporation__corporation_id")
    def _corporation__corporation_id(self, obj: CorporationAudit):
        return obj.corporation.corporation_id

    @admin.display(ordering="last_update_at", description=_("last update run"))
    def _last_update_at(self, obj: CorporationAudit):
        return naturaltime(obj.last_update_at) if obj.last_update_at else "-"

    # pylint: disable=unused-argument
    def has_add_permission(self, request):
        return False

    # pylint: disable=unused-argument
    def has_change_permission(self, request, obj=None):
        return False

    @admin.action(description=_("Force update selected corporations"))
    def force_update(self, request, queryset):
        """Force update of selected corporations."""
        count = 0
        for corporation_audit in queryset:
            update_corporation.delay(
                corporation_pk=corporation_audit.pk, force_refresh=True
            )
            count += 1

        messages.success(
            request,
            _(
                f"Started force update for {count} corporation(s). Updates will run in the background."
            ),
        )


class CharacterUpdateStatusAdminInline(admin.TabularInline):
    model = CharacterUpdateStatus
    fields = (
        "section",
        "_is_enabled",
        "_is_success",
        "_is_token_ok",
        "error_message",
        "run_finished_at",
        "_run_duration",
        "update_finished_at",
        "_update_duration",
    )
    readonly_fields = (
        "_is_enabled",
        "_is_success",
        "_is_token_ok",
        "_run_duration",
        "_update_duration",
    )
    ordering = ["section"]

    # pylint: disable=unused-argument
    def has_add_permission(self, request, obj=None):
        return False

    # pylint: disable=unused-argument
    def has_change_permission(self, request, obj=None):
        return False

    # pylint: disable=unused-argument
    def has_delete_permission(self, request, obj=None):
        return False

    @admin.display(boolean=True)
    def _is_enabled(self, obj: CharacterUpdateStatus) -> bool:
        return obj.is_enabled

    @admin.display(boolean=True)
    def _is_success(self, obj: CharacterUpdateStatus) -> bool:
        if not obj.is_enabled:
            return None
        return obj.is_success

    @admin.display(boolean=True)
    def _is_token_ok(self, obj: CharacterUpdateStatus) -> bool:
        return not obj.has_token_error

    @admin.display
    def _run_duration(self, obj: CharacterUpdateStatus) -> float:
        return self._calc_duration(obj.last_run_at, obj.last_run_finished_at)

    @admin.display
    def _update_duration(self, obj: CharacterUpdateStatus) -> float:
        return self._calc_duration(obj.last_update_at, obj.last_update_finished_at)

    @staticmethod
    def _calc_duration(
        started_at: timezone.datetime, finished_at: timezone.datetime
    ) -> str:
        if not started_at or not finished_at:
            return "-"
        return timesince(started_at, finished_at)


@admin.register(CharacterAudit)
class CharacterAuditAdmin(admin.ModelAdmin):
    list_display = (
        "_entity_pic",
        "_eve_character__character_name",
        "_last_update_at",
    )

    list_display_links = (
        "_entity_pic",
        "_eve_character__character_name",
    )

    list_select_related = ("eve_character",)

    ordering = ["eve_character__character_name"]

    search_fields = ["eve_character__character_name"]

    actions = [
        "delete_objects",
        "force_update",
    ]

    inlines = (CharacterUpdateStatusAdminInline,)

    def get_queryset(self, *args, **kwargs):
        qs = super().get_queryset(*args, **kwargs)
        return qs.prefetch_related("ledger_update_status").annotate(
            last_update_at=Max(
                "ledger_update_status__last_run_finished_at",
                filter=~Q(ledger_update_status__section="wallet_journal"),
            )
        )

    @admin.display(description="")
    def _entity_pic(self, obj: CharacterAudit):
        eve_id = obj.eve_character.character_id
        return format_html(
            '<img src="{}" class="img-circle">',
            eveimageserver._eve_entity_image_url("character", eve_id, 32),
        )

    @admin.display(
        description="Character Name", ordering="eve_character__character_name"
    )
    def _eve_character__character_name(self, obj: CharacterAudit):
        return obj.eve_character.character_name

    @admin.display(ordering="last_update_at", description=_("last update run"))
    def _last_update_at(self, obj: CharacterAudit):
        return naturaltime(obj.last_update_at) if obj.last_update_at else "-"

    # pylint: disable=unused-argument
    def has_add_permission(self, request):
        return False

    # pylint: disable=unused-argument
    def has_change_permission(self, request, obj=None):
        return False

    @admin.action(description=_("Force update selected characters"))
    def force_update(self, request, queryset):
        """Force update of selected characters."""
        count = 0
        for character_audit in queryset:
            update_character.delay(character_pk=character_audit.pk, force_refresh=True)
            count += 1

        messages.success(
            request,
            _(
                f"Started force update for {count} character(s). Updates will run in the background."
            ),
        )



"""
Character Audit
"""

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import redirect
from django.utils.translation import gettext_lazy as trans

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from esi.decorators import token_required

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__, tasks
from ledger.models.characteraudit import CharacterAudit

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@token_required(scopes=CharacterAudit.get_esi_scopes())
@permission_required("ledger.basic_access")
def add_char(request, token):
    char, _ = CharacterAudit.objects.update_or_create(
        eve_character=EveCharacter.objects.get_character_by_id(token.character_id),
        defaults={
            "active": True,
            "character_name": token.character_name,
        },
    )
    tasks.update_character.apply_async(
        args=[char.pk], kwargs={"force_refresh": True}, priority=6
    )

    msg = trans("{character_name} successfully added/updated to Ledger").format(
        character_name=char.eve_character.character_name,
    )
    messages.info(request, msg)
    return redirect(
        "ledger:character_ledger", character_id=char.eve_character.character_id
    )



# Standard Library
from typing import Any

# Third Party
from ninja import NinjaAPI

# Django
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.authentication.models import CharacterOwnership, UserProfile
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.api import schema
from ledger.api.helpers import (
    get_all_corporations_from_alliance,
    get_character_or_none,
    get_corporation,
)
from ledger.models.characteraudit import CharacterAudit
from ledger.models.corporationaudit import CorporationAudit

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class LedgerAdminApiEndpoints:
    tags = ["LedgerAdmin"]

    # pylint: disable=too-many-statements
    def __init__(self, api: NinjaAPI):
        @api.get(
            "character/overview/",
            response={200: list[schema.CharacterAdmin], 403: str},
            tags=self.tags,
        )
        def get_character_overview(request):
            chars_visible = CharacterAudit.objects.visible_to(request.user)

            if chars_visible is None:
                return 403, "Permission Denied"

            chars_ids = chars_visible.values_list(
                "eve_character__character_id", flat=True
            )

            users_char_ids = UserProfile.objects.filter(
                main_character__isnull=False, main_character__character_id__in=chars_ids
            )

            character_dict = {}

            for character in users_char_ids:
                # pylint: disable=broad-exception-caught
                try:
                    character_dict[character.main_character.character_id] = {
                        "character_id": character.main_character.character_id,
                        "character_name": character.main_character.character_name,
                        "corporation_id": character.main_character.corporation_id,
                        "corporation_name": character.main_character.corporation_name,
                    }
                except AttributeError:
                    continue

            output = []
            output.append({"character": character_dict})

            return output

        @api.get(
            "planetary/overview/",
            response={200: list[schema.CharacterAdmin], 403: str},
            tags=self.tags,
        )
        def get_planetary_overview(request):
            chars_ids = CharacterAudit.objects.visible_eve_characters(
                request.user
            ).values_list("character_id", flat=True)

            users_char_ids = UserProfile.objects.filter(
                main_character__isnull=False, main_character__character_id__in=chars_ids
            )

            if not chars_ids:
                return 403, "Permission Denied"

            character_dict = {}

            for character in users_char_ids:
                # pylint: disable=broad-exception-caught
                try:
                    character_dict[character.main_character.character_id] = {
                        "character_id": character.main_character.character_id,
                        "character_name": character.main_character.character_name,
                        "corporation_id": character.main_character.corporation_id,
                        "corporation_name": character.main_character.corporation_name,
                    }
                except Exception:
                    continue

            output = []
            output.append({"character": character_dict})

            return output

        @api.get(
            "corporation/overview/",
            response={200: list[schema.CorporationAdmin], 403: str},
            tags=self.tags,
        )
        def get_corporation_overview(request):
            corporations = CorporationAudit.objects.visible_to(request.user)

            if corporations is None:
                return 403, "Permission Denied"

            corporation_dict = {}

            for corporation in corporations:
                # pylint: disable=broad-exception-caught
                try:
                    corporation_dict[corporation.corporation.corporation_id] = {
                        "corporation_id": corporation.corporation.corporation_id,
                        "corporation_name": corporation.corporation.corporation_name,
                    }
                except Exception:
                    continue

            output = []
            output.append({"corporation": corporation_dict})

            return output

        @api.get(
            "alliance/overview/",
            response={200: list[schema.AllianceAdmin], 403: str},
            tags=self.tags,
        )
        def get_alliance_overview(request):
            corporations = CorporationAudit.objects.visible_to(request.user)

            if corporations is None:
                return 403, "Permission Denied"

            alliance_dict = {}

            for corporation in corporations:
                # pylint: disable=broad-exception-caught
                try:
                    alliance_dict[corporation.corporation.alliance.alliance_id] = {
                        "alliance_id": corporation.corporation.alliance.alliance_id,
                        "alliance_name": corporation.corporation.alliance.alliance_name,
                    }
                except Exception:
                    continue
            output = []
            output.append({"alliance": alliance_dict})

            return output

        @api.get(
            "character/{character_id}/view/dashboard/",
            response={200: Any, 403: str},
            tags=self.tags,
        )
        def get_character_dashboard(request, character_id: int):
            perm, character = get_character_or_none(request, character_id)

            if not perm:
                return 403, "Permission Denied"
            if perm is None:
                return 403, "Character not found"

            linked_characters_ids = character.alts.values_list(
                "character_id", flat=True
            )

            characters = CharacterAudit.objects.filter(
                eve_character__character_id__in=linked_characters_ids
            )

            auth_characters = len(linked_characters_ids)
            active_characters = characters.filter(active=True).count()
            inactive_characters = characters.filter(active=False).count()
            missing_characters = (
                auth_characters - active_characters - inactive_characters
            )

            status_msg = None
            status_issues = None
            issues = []
            for char in characters:
                if char.ledger_update_status.filter(is_success=False).exists():
                    issues.append(char.eve_character.character_name)

            if issues:
                status_msg = _("Please re-register characters with issues")
                status_issues = ", ".join(issues)

            output = {
                "dashboard": "Character Dashboard",
                "status": status_msg,
                "status_issues": status_issues,
                "statistics": "Character Statistics",
                "auth_characters": auth_characters,
                "active_characters": f"{active_characters} / {auth_characters}",
                "inactive_characters": inactive_characters,
                "missing_characters": missing_characters,
            }

            return output

        @api.get(
            "corporation/{corporation_id}/view/dashboard/",
            response={200: Any, 403: str},
            tags=self.tags,
        )
        def get_corporation_dashboard(request, corporation_id: int):
            perm, corporation = get_corporation(request, corporation_id)

            if not perm:
                return 403, "Permission Denied"
            if perm is None:
                return 403, "Corporation not found"

            auth_corp = EveCorporationInfo.objects.get(
                corporation_id=corporation.corporation.corporation_id
            )

            corp_characters = CharacterOwnership.objects.filter(
                character__corporation_id=corporation_id
            )

            auth_characters = auth_corp.member_count
            active_characters = corp_characters.count()
            missing_characters = auth_characters - active_characters

            output = {
                "dashboard": "Corporation Dashboard",
                "status": "Missing Characters are not impleted yet",
                "statistics": "Corporation Statistics",
                "auth_characters": auth_characters,
                "active_characters": f"{active_characters} / {auth_characters}",
                "missing_characters": missing_characters,
            }

            return output

        @api.get(
            "alliance/{alliance_id}/view/dashboard/",
            response={200: Any, 403: str},
            tags=self.tags,
        )
        def get_alliance_dashboard(request, alliance_id: int):
            perm, corporations = get_all_corporations_from_alliance(
                request, alliance_id
            )

            if not perm:
                return 403, "Permission Denied"
            if perm is None:
                return 403, "Alliance not found."

            all_corporations = EveCorporationInfo.objects.filter(
                alliance__alliance_id=alliance_id
            )
            corp_audit_ids = corporations.values_list(
                "corporation__corporation_id", flat=True
            )
            missing_corporations = all_corporations.exclude(
                corporation_id__in=corp_audit_ids
            )

            active_corporations = corporations.count()

            output = {
                "dashboard": "Alliance Dashboard",
                "statistics": "Alliance Statistics",
                "auth_corporations": all_corporations.count(),
                "active_corporations": f"{active_corporations} / {all_corporations.count()}",
                "missing_corporations": missing_corporations.count(),
            }

            return output



"""PvE Views"""

# Standard Library
import json
from decimal import Decimal

# Django
from django.core.handlers.wsgi import WSGIRequest
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.helpers.core import LedgerCore
from ledger.models.characteraudit import (
    CharacterAudit,
    CharacterMiningLedger,
    CharacterWalletJournalEntry,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharacterData(LedgerCore):
    """Class to hold character data for the ledger."""

    # pylint: disable=too-many-positional-arguments
    def __init__(
        self,
        request: WSGIRequest,
        character: CharacterAudit,
        year=None,
        month=None,
        day=None,
    ):
        LedgerCore.__init__(self, year, month, day)
        self.request = request
        self.character = character
        self.alts_ids = self.get_alt_ids

    @property
    def get_alt_ids(self):
        return self.character.alts.values_list("character_id", flat=True)

    @property
    def is_old_ess(self):
        """
        Compatibility check for old ESS income calculation.
        Since Swagger ESI has added ESS Ref Type to the Character Journal Endpoint
        """
        try:
            if self.month is None and self.year is None:
                return False
            if self.year >= 2025 and self.month >= 6:
                return False
        except TypeError:
            return True
        return True

    def setup_ledger(self, character: CharacterAudit):
        """Setup the Ledger Data for the Character."""

        # Show Card Template for Character Ledger
        if self.request.GET.get("single", False):
            self.ledger_type = "single"

        # Get All Journal Entries for the Character and its Alts for Details View
        if self.request.GET.get("all", False):
            self.journal = CharacterWalletJournalEntry.objects.filter(
                self.filter_date,
                character__eve_character__character_id__in=self.alts_ids,
            )
            self.mining = CharacterMiningLedger.objects.filter(
                self.filter_date,
                character__eve_character__character_id__in=self.alts_ids,
            )
        else:
            # Get Journal Entries for the Character and its Alts
            self.journal = character.ledger_character_journal.filter(self.filter_date)
            self.mining = character.ledger_character_mining.filter(self.filter_date)

    def generate_ledger_data(self) -> dict:
        """Generate the ledger data for the character and its alts."""
        # Only show the character if 'single' is set in the request
        if self.request.GET.get("single", False):
            characters = CharacterAudit.objects.filter(
                eve_character__character_id=self.character.eve_character.character_id
            )
            character = characters.first()
            character_data = self._create_character_data(character=character)
            ledger = character_data
        else:
            ledger = []
            characters = CharacterAudit.objects.filter(
                eve_character__character_id__in=self.alts_ids
            ).select_related("eve_character")
            for character in characters:
                character_data = self._create_character_data(character=character)
                if character_data:
                    ledger.append(character_data)

        # Generate the totals for the ledger
        totals = self._calculate_totals(ledger)

        # Evaluate the existing years for the view
        existing_years = (
            CharacterWalletJournalEntry.objects.filter(character__in=characters)
            .exclude(date__year__isnull=True)
            .values_list("date__year", flat=True)
            .order_by("-date__year")
            .distinct()
        )

        # Create the ratting bar for the view
        self.create_rattingbar(
            is_old_ess=self.is_old_ess,
            character_ids=characters.values_list(
                "eve_character__character_id", flat=True
            ),
        )

        context = {
            "title": f"Character Ledger - {self.character.eve_character.character_name}",
            "character_id": self.character.eve_character.character_id,
            "billboard": json.dumps(self.billboard.dict.asdict()),
            "ledger": ledger,
            "years": list(existing_years),
            "totals": totals,
            "view": self.create_view_data(
                viewname="character_details",
                character_id=self.character.eve_character.character_id,
            ),
            "is_old_ess": self.is_old_ess,
        }
        return context

    def _create_character_data(
        self,
        character: CharacterAudit,
    ):
        """Create a dictionary with character data and update billboard/ledger."""
        self.setup_ledger(character)

        # If no journal or mining data exists, return None
        if not self.journal.exists() and not self.mining.exists():
            return None

        bounty = self.journal.aggregate_bounty()
        ess = (
            self.journal.aggregate_bounty() * Decimal(0.667)
            if self.is_old_ess
            else self.journal.aggregate_ess()
        )
        mining_val = self.mining.aggregate_mining()
        costs = self.journal.aggregate_costs(second_party=self.alts_ids)
        miscellaneous = self.journal.aggregate_miscellaneous(first_party=self.alts_ids)
        total = sum(
            [
                bounty,
                ess,
                costs,
                miscellaneous,
            ]
        )

        # If total is 0, we do not need to create a character data entry
        if int(total) == 0:
            return None

        update_states = {}

        for status in character.ledger_update_status.all():
            update_states[status.section] = {
                "is_success": status.is_success,
                "last_update_finished_at": status.last_update_finished_at,
                "last_run_finished_at": status.last_run_finished_at,
            }

        char_data = {
            "character": character,
            "ledger": {
                "bounty": bounty,
                "ess": ess,
                "mining": mining_val,
                "costs": costs,
                "miscellaneous": miscellaneous,
                "total": total,
            },
            "update_states": update_states,
            "single_url": self.create_url(
                viewname="character_ledger",
                character_id=character.eve_character.character_id,
            ),
            "details_url": self.create_url(
                viewname="character_details",
                character_id=character.eve_character.character_id,
            ),
        }

        # Create the chord data for the billboard
        self.billboard.chord_add_data(
            chord_from=character.eve_character.character_name,
            chord_to=_("Bounty"),
            value=bounty,
        )
        self.billboard.chord_add_data(
            chord_from=character.eve_character.character_name,
            chord_to=_("ESS"),
            value=ess,
        )
        self.billboard.chord_add_data(
            chord_from=character.eve_character.character_name,
            chord_to=_("Mining"),
            value=mining_val,
        )
        self.billboard.chord_add_data(
            chord_from=character.eve_character.character_name,
            chord_to=_("Miscellaneous"),
            value=miscellaneous,
        )
        self.billboard.chord_add_data(
            chord_from=character.eve_character.character_name,
            chord_to=_("Costs"),
            value=abs(costs),
        )

        return char_data

    def create_rattingbar(self, character_ids: list = None, is_old_ess: bool = False):
        """Create the ratting bar for the view."""
        if not character_ids:
            return

        rattingbar_timeline = self.billboard.create_timeline(
            CharacterWalletJournalEntry.objects.filter(
                self.filter_date,
                character__eve_character__character_id__in=character_ids,
            )
        )
        rattingbar_mining_timeline = self.billboard.create_timeline(
            CharacterMiningLedger.objects.filter(
                self.filter_date,
                character__eve_character__character_id__in=character_ids,
            )
        )

        rattingbar = (
            rattingbar_timeline.annotate_bounty_income()
            .annotate_ess_income()
            .annotate_miscellaneous_with_exclude(exclude=self.alts_ids)
        )
        rattingbar_mining = rattingbar_mining_timeline.annotate_mining(with_period=True)
        self.billboard.create_or_update_results(rattingbar, is_old_ess=is_old_ess)
        self.billboard.add_category(rattingbar_mining, category="mining")
        self._build_xy_chart(title=_("Ratting Bar"))



# Standard Library
from typing import TYPE_CHECKING

# Django
from django.db import models, transaction
from django.db.models import DecimalField, Q, Sum, Value
from django.db.models.functions import Coalesce
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.decorators import log_timing
from ledger.errors import DatabaseError, NotModifiedError
from ledger.helpers.etag import etag_results
from ledger.helpers.ref_type import RefTypeManager
from ledger.models.general import EveEntity
from ledger.providers import esi

if TYPE_CHECKING:
    # AA Ledger
    from ledger.models.corporationaudit import (
        CorporationAudit,
        CorporationWalletDivision,
    )

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CorporationWalletQuerySet(models.QuerySet):
    # pylint: disable=duplicate-code
    def annotate_bounty_income(
        self,
    ) -> models.QuerySet:
        return self.annotate(
            bounty_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.BOUNTY_PRIZES, amount__gt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_ess_income(self) -> models.QuerySet:
        return self.annotate(
            ess_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.ESS_TRANSFER, amount__gt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    # pylint: disable=duplicate-code
    def annotate_miscellaneous(self) -> models.QuerySet:
        return self.annotate(
            miscellaneous=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__gt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_costs(self) -> models.QuerySet:
        return self.annotate(
            costs=Coalesce(
                Sum(
                    "amount",
                    filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def aggregate_bounty(self) -> dict:
        """Aggregate bounty income."""
        return self.filter(ref_type__in=RefTypeManager.BOUNTY_PRIZES).aggregate(
            total_bounty=Coalesce(Sum("amount"), Value(0), output_field=DecimalField())
        )["total_bounty"]

    def aggregate_ess(self) -> dict:
        """Aggregate ESS income."""
        return self.filter(ref_type__in=RefTypeManager.ESS_TRANSFER).aggregate(
            total_ess=Coalesce(Sum("amount"), Value(0), output_field=DecimalField())
        )["total_ess"]

    def aggregate_miscellaneous(self) -> dict:
        """Aggregate miscellaneous income (nur positive Beträge)."""
        return self.filter(
            ref_type__in=RefTypeManager.all_ref_types(), amount__gt=0
        ).aggregate(
            total_misc=Coalesce(Sum("amount"), Value(0), output_field=DecimalField())
        )[
            "total_misc"
        ]

    def aggregate_costs(self) -> dict:
        """Aggregate costs."""
        return self.filter(
            ref_type__in=RefTypeManager.all_ref_types(), amount__lt=0
        ).aggregate(
            total_costs=Coalesce(Sum("amount"), Value(0), output_field=DecimalField())
        )[
            "total_costs"
        ]

    # pylint: disable=too-many-positional-arguments, duplicate-code
    def aggregate_ref_type(
        self,
        ref_type: list,
        first_party=None,
        second_party=None,
        exclude=None,
        income: bool = False,
    ) -> dict:
        """Aggregate income by ref_type."""
        qs = self.filter(ref_type__in=ref_type)
        if first_party is not None:
            if isinstance(first_party, int):
                first_party = [first_party]
            qs = qs.filter(first_party__in=first_party)

        if second_party is not None:
            if isinstance(second_party, int):
                second_party = [second_party]
            qs = qs.filter(second_party__in=second_party)

        if exclude is not None:
            if isinstance(exclude, int):
                exclude = [exclude]
            qs = qs.exclude(first_party__in=exclude)

        if income:
            qs = qs.filter(amount__gt=0)
        else:
            qs = qs.filter(amount__lt=0)

        return qs.aggregate(
            total=Coalesce(Sum("amount"), Value(0), output_field=DecimalField())
        )["total"]


class CorporationWalletManagerBase(models.Manager):
    @log_timing(logger)
    def update_or_create_esi(
        self, corporation: "CorporationAudit", force_refresh: bool = False
    ) -> None:
        """Update or Create a wallet journal entry from ESI data."""
        return corporation.update_section_if_changed(
            section=corporation.UpdateSection.WALLET_JOURNAL,
            fetch_func=self._fetch_esi_data,
            force_refresh=force_refresh,
        )

    def _fetch_esi_data(
        self, corporation: "CorporationAudit", force_refresh: bool = False
    ) -> None:
        """Fetch wallet journal entries from ESI data."""
        # AA Ledger
        # pylint: disable=import-outside-toplevel
        from ledger.models.corporationaudit import CorporationWalletDivision

        req_scopes = [
            "esi-wallet.read_corporation_wallets.v1",
            "esi-characters.read_corporation_roles.v1",
        ]
        req_roles = ["CEO", "Director", "Accountant", "Junior_Accountant"]

        token = corporation.get_token(scopes=req_scopes, req_roles=req_roles)

        divisions = CorporationWalletDivision.objects.filter(corporation=corporation)

        # Get the count of divisions to track not modified
        division_count = divisions.count()
        not_modified = 0

        for division in divisions:
            current_page = 1
            total_pages = 1
            while current_page <= total_pages:
                journal_items_ob = esi.client.Wallet.get_corporations_corporation_id_wallets_division_journal(
                    corporation_id=corporation.corporation.corporation_id,
                    division=division.division_id,
                    page=current_page,
                    token=token.valid_access_token(),
                )

                journal_items_ob.request_config.also_return_response = True
                __, headers = journal_items_ob.result()
                total_pages = int(headers.headers.get("X-Pages", 1))

                logger.debug(
                    "Fetching Journal Items for %s - Division: %s - Page: %s/%s",
                    corporation.corporation.corporation_name,
                    division.division_id,
                    current_page,
                    total_pages,
                )

                try:
                    journal_items = etag_results(
                        journal_items_ob,
                        token,
                        force_refresh=force_refresh,
                    )
                except NotModifiedError:
                    not_modified += 1
                    logger.debug(
                        "NotModifiedError: %s - Division: %s - Page: %s",
                        corporation.corporation.corporation_name,
                        division.division_id,
                        current_page,
                    )
                    current_page += 1
                    continue

                self._update_or_create_objs(division, journal_items)
                current_page += 1

        # Ensure only raise NotModifiedError if all divisions returned NotModified
        if not_modified == division_count:
            raise NotModifiedError()

    @transaction.atomic()
    def _update_or_create_objs(
        self,
        division: "CorporationWalletDivision",
        objs: list,
    ) -> None:
        """Update or Create wallet journal entries from objs data."""
        _new_names = []
        _current_journal = set(
            list(
                self.filter(division=division)
                .order_by("-date")
                .values_list("entry_id", flat=True)[:20000]
            )
        )
        _current_eve_ids = set(
            list(EveEntity.objects.all().values_list("eve_id", flat=True))
        )

        items = []
        # pylint: disable=duplicate-code
        for item in objs:
            if item.get("id") not in _current_journal:
                if item.get("second_party_id") not in _current_eve_ids:
                    _new_names.append(item.get("second_party_id"))
                    _current_eve_ids.add(item.get("second_party_id"))
                if item.get("first_party_id") not in _current_eve_ids:
                    _new_names.append(item.get("first_party_id"))
                    _current_eve_ids.add(item.get("first_party_id"))

                wallet_item = self.model(  # pylint: disable=duplicate-code
                    division=division,
                    amount=item.get("amount"),
                    balance=item.get("balance"),
                    context_id=item.get("context_id"),
                    context_id_type=item.get("context_id_type"),
                    date=item.get("date"),
                    description=item.get("description"),
                    first_party_id=item.get("first_party_id"),
                    entry_id=item.get("id"),
                    reason=item.get("reason"),
                    ref_type=item.get("ref_type"),
                    second_party_id=item.get("second_party_id"),
                    tax=item.get("tax"),
                    tax_receiver_id=item.get("tax_receiver_id"),
                )

                items.append(wallet_item)

        created_names = EveEntity.objects.create_bulk_from_esi(_new_names)

        if created_names:
            self.bulk_create(items)
        else:
            raise DatabaseError("DB Fail")

        logger.debug(
            "Created %s Journal Entries for %s",
            len(items),
            division.corporation.corporation_name,
        )


CorporationWalletManager = CorporationWalletManagerBase.from_queryset(
    CorporationWalletQuerySet
)


class CorporationDivisionQuerySet(models.QuerySet):
    pass


class CorporationDivisionManagerBase(models.Manager):
    @log_timing(logger)
    def update_or_create_esi(
        self, corporation: "CorporationAudit", force_refresh: bool = False
    ) -> None:
        """Update or Create a division entry from ESI data."""
        return corporation.update_section_if_changed(
            section=corporation.UpdateSection.WALLET_DIVISION,
            fetch_func=self._fetch_esi_data,
            force_refresh=force_refresh,
        )

    @log_timing(logger)
    def update_or_create_esi_names(
        self, corporation: "CorporationAudit", force_refresh: bool = False
    ) -> None:
        """Update or Create a division entry from ESI data."""
        return corporation.update_section_if_changed(
            section=corporation.UpdateSection.WALLET_DIVISION_NAMES,
            fetch_func=self._fetch_esi_data_names,
            force_refresh=force_refresh,
        )

    def _fetch_esi_data_names(
        self, corporation: "CorporationAudit", force_refresh: bool = False
    ) -> None:
        """Fetch division entries from ESI data."""
        req_scopes = [
            "esi-corporations.read_divisions.v1",
        ]
        req_roles = ["CEO", "Director"]

        token = corporation.get_token(scopes=req_scopes, req_roles=req_roles)

        division_obj = esi.client.Corporation.get_corporations_corporation_id_divisions(
            corporation_id=corporation.corporation.corporation_id,
        )

        division_names = etag_results(division_obj, token, force_refresh=force_refresh)

        self._update_or_create_objs_division(corporation, division_names)

    def _fetch_esi_data(
        self, corporation: "CorporationAudit", force_refresh: bool = False
    ) -> None:
        """Fetch division entries from ESI data."""
        req_scopes = [
            "esi-wallet.read_corporation_wallets.v1",
            "esi-characters.read_corporation_roles.v1",
            "esi-corporations.read_divisions.v1",
        ]
        req_roles = ["CEO", "Director", "Accountant", "Junior_Accountant"]

        token = corporation.get_token(scopes=req_scopes, req_roles=req_roles)

        divisions_items_obj = esi.client.Wallet.get_corporations_corporation_id_wallets(
            corporation_id=corporation.corporation.corporation_id
        )

        division_balances = etag_results(
            divisions_items_obj, token, force_refresh=force_refresh
        )
        self._update_or_create_objs(corporation, division_balances)

    @transaction.atomic()
    def _update_or_create_objs_division(
        self,
        corporation: "CorporationAudit",
        objs: list,
    ) -> None:
        """Update or Create division entries from objs data."""
        for division in objs.get("wallet"):
            if division.get("division") == 1:
                name = _("Master Wallet")
            else:
                name = division.get("name", _("Unknown"))

            obj, created = self.get_or_create(
                corporation=corporation,
                division_id=division.get("division"),
                defaults={"balance": 0, "name": name},
            )
            if not created:
                obj.name = name
                obj.save()

    @transaction.atomic()
    def _update_or_create_objs(
        self,
        corporation: "CorporationAudit",
        objs: list,
    ) -> None:
        """Update or Create division entries from objs data."""
        for division in objs:
            obj, created = self.get_or_create(
                corporation=corporation,
                division_id=division.get("division"),
                defaults={
                    "balance": division.get("balance"),
                    "name": _("Unknown"),
                },
            )

            if not created:
                obj.balance = division.get("balance")
                obj.save()


CorporationDivisionManager = CorporationDivisionManagerBase.from_queryset(
    CorporationDivisionQuerySet
)



# Generated by Django 4.2.20 on 2025-05-04 21:51

# Django
import django.db.models.deletion
from django.db import migrations, models


def set_character_from_planet(apps, schema_editor):
    CharacterPlanetDetails = apps.get_model("ledger", "CharacterPlanetDetails")
    CharacterPlanet = apps.get_model("ledger", "CharacterPlanet")

    for detail in CharacterPlanetDetails.objects.all():
        if detail.planet_id:
            try:
                planet = CharacterPlanet.objects.get(id=detail.planet_id)
                detail.character = planet.character
                detail.save()
            except CharacterPlanet.DoesNotExist:
                detail.delete()
                detail.save()
        else:
            detail.delete()
            detail.save()


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0009_remove_characteraudit_last_update_mining_and_more"),
    ]

    operations = [
        migrations.RunPython(set_character_from_planet),
        migrations.AlterField(
            model_name="characterplanetdetails",
            name="character",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ledger_character_planet_details",
                to="ledger.characteraudit",
            ),
        ),
    ]



# Standard Library
from datetime import datetime
from typing import Any

# Third Party
from ninja import Schema


class EveName(Schema):
    id: int
    name: str
    cat: str | None = None


class CharacterPlanet(Schema):
    character_id: int | None = None
    character_name: str | None = None
    planet: str | None = None
    planet_id: int | None = None
    upgrade_level: int | None = None
    num_pins: int | None = None
    last_update: datetime | None = None


class CharacterAdmin(Schema):
    character: dict | None = None


class CorporationAdmin(Schema):
    corporation: dict | None = None


class AllianceAdmin(Schema):
    alliance: dict | None = None


class Ledger(Schema):
    ratting: list | None = None
    billboard: Any | None = None
    total: dict | None = None



# Standard Library
import json
from pathlib import Path

# Alliance Auth (External Libs)
from app_utils.esi_testing import EsiClientStub, EsiEndpoint


def load_test_data():
    file_path = Path(__file__).parent / "esi.json"
    with file_path.open("r", encoding="utf-8") as fp:
        return json.load(fp)


_esi_data = load_test_data()

_endpoints = [
    EsiEndpoint(
        "Character",
        "get_characters_character_id_roles",
        "character_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Wallet",
        "get_characters_character_id_wallet_journal",
        "character_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Wallet",
        "get_characters_character_id_wallet",
        "character_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Corporation",
        "get_corporations_corporation_id_divisions",
        "corporation_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Wallet",
        "get_corporations_corporation_id_wallets",
        "corporation_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Wallet",
        "get_corporations_corporation_id_wallets_division_journal",
        "corporation_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Industry",
        "get_characters_character_id_mining",
        "character_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Planetary_Interaction",
        "get_characters_character_id_planets",
        "character_id",
        needs_token=False,
    ),
    EsiEndpoint(
        "Planetary_Interaction",
        "get_characters_character_id_planets_planet_id",
        ("character_id", "planet_id"),
        needs_token=False,
    ),
    EsiEndpoint(
        "Universe",
        "post_universe_names",
        "ids",
        needs_token=False,
    ),
]

esi_client_stub = EsiClientStub(_esi_data, endpoints=_endpoints)
esi_client_error_stub = EsiClientStub(_esi_data, endpoints=_endpoints, http_error=502)



"""
General Model
"""

# Standard Library
import datetime
from dataclasses import dataclass
from typing import Any, NamedTuple

# Django
from django.core.validators import MinValueValidator
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__, app_settings
from ledger.managers.general_manager import EveEntityManager

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

# Permission Manager


class General(models.Model):
    """A model defining commonly used properties and methods for Ledger."""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app, Ledger."),
            ("advanced_access", "Can access Corporation and Alliance Ledger."),
            ("manage_access", "Can manage Ledger."),
        )


# EvE Entity Model - Store all Chars, Corps, Allys
class EveEntity(models.Model):
    """An Eve entity like a corporation or a character"""

    CATEGORY_ALLIANCE = "alliance"
    CATEGORY_CHARACTER = "character"
    CATEGORY_CORPORATION = "corporation"

    CATEGORY_CHOICES = (
        (CATEGORY_ALLIANCE, "Alliance"),
        (CATEGORY_CORPORATION, "Corporation"),
        (CATEGORY_CHARACTER, "Character"),
    )

    eve_id = models.IntegerField(
        primary_key=True, validators=[MinValueValidator(0)], verbose_name=_("ID")
    )
    category = models.CharField(
        max_length=32, choices=CATEGORY_CHOICES, verbose_name=_("Category")
    )
    name = models.CharField(max_length=254, verbose_name=_("Name"))

    # optionals for character/corp
    corporation = models.ForeignKey(
        "EveEntity",
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="corp",
    )
    alliance = models.ForeignKey(
        "EveEntity",
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="alli",
    )
    last_update = models.DateTimeField(auto_now=True)

    objects = EveEntityManager()

    def __str__(self) -> str:
        return str(self.name)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.eve_id}, category='{self.category}', "
            f"name='{self.name}')"
        )

    @property
    def is_alliance(self) -> bool:
        """Return True if entity is an alliance, else False."""
        return self.category == self.CATEGORY_ALLIANCE

    @property
    def is_corporation(self) -> bool:
        """Return True if entity is an corporation, else False."""
        return self.category == self.CATEGORY_CORPORATION

    @property
    def is_character(self) -> bool:
        """Return True if entity is a character, else False."""
        return self.category == self.CATEGORY_CHARACTER

    def icon_url(self, size=128) -> str:
        """Url to an icon image for this organization."""
        if self.category == self.CATEGORY_ALLIANCE:
            return EveAllianceInfo.generic_logo_url(self.eve_id, size=size)

        if self.category == self.CATEGORY_CORPORATION:
            return EveCorporationInfo.generic_logo_url(self.eve_id, size=size)

        if self.category == self.CATEGORY_CHARACTER:
            return EveCharacter.generic_portrait_url(self.eve_id, size=size)

        raise NotImplementedError(
            f"Avatar URL not implemented for category {self.category}"
        )

    def needs_update(self):
        return self.last_update + datetime.timedelta(days=7) < timezone.now()

    class Meta:
        default_permissions = ()


class UpdateSectionResult(NamedTuple):
    """A result of an attempted section update."""

    is_changed: bool | None
    is_updated: bool
    data: Any = None


@dataclass(frozen=True)
class _NeedsUpdate:
    """An Object to track if an update is needed."""

    section_map: dict[str, bool]

    def __bool__(self) -> bool:
        """Check if any section needs an update."""
        return any(self.section_map.values())

    def for_section(self, section: str) -> bool:
        """Check if an update is needed for a specific section."""
        return self.section_map.get(section, False)


class UpdateStatus(models.Model):
    """A Model to track the status of the last update."""

    is_success = models.BooleanField(default=None, null=True, db_index=True)
    error_message = models.TextField()
    has_token_error = models.BooleanField(default=False)

    last_run_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last run has been started at this time",
    )
    last_run_finished_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last run has been successful finished at this time",
    )
    last_update_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last update has been started at this time",
    )
    last_update_finished_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last update has been successful finished at this time",
    )

    class Meta:
        abstract = True
        default_permissions = ()

    def need_update(self) -> bool:
        """Check if the update is needed."""
        if not self.is_success or not self.last_update_finished_at:
            needs_update = True
        else:
            section_time_stale = app_settings.LEDGER_STALE_TYPES.get(self.section, 60)
            stale = timezone.now() - timezone.timedelta(minutes=section_time_stale)

            try:
                needs_update = self.last_update_finished_at <= stale
            except AttributeError:
                needs_update = True

        if needs_update and self.has_token_error:
            logger.info(
                "%s: Ignoring update because of token error, section: %s",
                self.corporation,
                self.section,
            )
            needs_update = False

        return needs_update

    def reset(self) -> None:
        """Reset this update status."""
        self.is_success = None
        self.error_message = ""
        self.has_token_error = False
        self.last_run_at = timezone.now()
        self.last_run_finished_at = None
        self.save()



"""App URLs"""

# Django
from django.urls import path, re_path

# AA Ledger
from ledger.api import api
from ledger.views.alliance import alliance_ledger
from ledger.views.alliance.add_ally import add_ally
from ledger.views.character import character_ledger, planetary
from ledger.views.character.add_char import add_char
from ledger.views.corporation import corporation_ledger
from ledger.views.corporation.add_corp import add_corp

# AA Example App
from ledger.views.index import admin, index

app_name: str = "ledger"

urlpatterns = [
    path("", index, name="index"),
    path("admin/", admin, name="admin"),
    # -- Character Audit
    path("character/add/", add_char, name="add_char"),
    path(
        "character/delete/<int:character_id>/",
        character_ledger.character_delete,
        name="delete_char",
    ),
    # -- Corporation Audit
    path("corporation/add/", add_corp, name="add_corp"),
    path(
        "corporation/delete/<int:corporation_id>/",
        corporation_ledger.corporation_delete,
        name="delete_corp",
    ),
    # -- -- Alliance Ledger
    path(
        "alliance/",
        alliance_ledger.alliance_ledger_index,
        name="alliance_ledger_index",
    ),
    path(
        "alliance/<int:alliance_id>/",
        alliance_ledger.alliance_ledger,
        name="alliance_ledger",
    ),
    path(
        "alliance/<int:alliance_id>/<int:year>/",
        alliance_ledger.alliance_ledger,
        name="alliance_ledger",
    ),
    path(
        "alliance/<int:alliance_id>/<int:year>/<int:month>/",
        alliance_ledger.alliance_ledger,
        name="alliance_ledger",
    ),
    path(
        "alliance/<int:alliance_id>/<int:year>/<int:month>/<int:day>/",
        alliance_ledger.alliance_ledger,
        name="alliance_ledger",
    ),
    # -- -- Alliance Details
    path(
        "alliance/<int:alliance_id>/<int:year>/view/details/<int:entity_id>/",
        alliance_ledger.alliance_details,
        name="alliance_details",
    ),
    path(
        "alliance/<int:alliance_id>/<int:year>/<int:month>/view/details/<int:entity_id>/",
        alliance_ledger.alliance_details,
        name="alliance_details",
    ),
    path(
        "alliance/<int:alliance_id>/<int:year>/<int:month>/<int:day>/view/details/<int:entity_id>/",
        alliance_ledger.alliance_details,
        name="alliance_details",
    ),
    path(
        "alliance/<int:alliance_id>/view/administration/",
        alliance_ledger.alliance_administration,
        name="alliance_administration",
    ),
    path(
        "alliance/view/overview/",
        alliance_ledger.alliance_overview,
        name="alliance_overview",
    ),
    path("alliance/add/", add_ally, name="add_ally"),
    # -- -- Corporation Ledger
    path(
        "corporation/",
        corporation_ledger.corporation_ledger_index,
        name="corporation_ledger_index",
    ),
    path(
        "corporation/<int:corporation_id>/",
        corporation_ledger.corporation_ledger,
        name="corporation_ledger",
    ),
    path(
        "corporation/<int:corporation_id>/date/<int:year>/",
        corporation_ledger.corporation_ledger,
        name="corporation_ledger",
    ),
    path(
        "corporation/<int:corporation_id>/date/<int:year>/<int:month>/",
        corporation_ledger.corporation_ledger,
        name="corporation_ledger",
    ),
    path(
        "corporation/<int:corporation_id>/date/<int:year>/<int:month>/<int:day>/",
        corporation_ledger.corporation_ledger,
        name="corporation_ledger",
    ),
    path(
        "corporation/<int:corporation_id>/division/<int:division_id>/",
        corporation_ledger.corporation_ledger,
        name="corporation_ledger",
    ),
    path(
        "corporation/<int:corporation_id>/division/<int:division_id>/date/<int:year>/",
        corporation_ledger.corporation_ledger,
        name="corporation_ledger",
    ),
    path(
        "corporation/<int:corporation_id>/division/<int:division_id>/date/<int:year>/<int:month>/",
        corporation_ledger.corporation_ledger,
        name="corporation_ledger",
    ),
    path(
        "corporation/<int:corporation_id>/division/<int:division_id>/date/<int:year>/<int:month>/<int:day>/",
        corporation_ledger.corporation_ledger,
        name="corporation_ledger",
    ),
    # -- -- Corporation Details
    path(
        "corporation/<int:corporation_id>/view/details/<int:entity_id>/",
        corporation_ledger.corporation_details,
        name="corporation_details",
    ),
    path(
        "corporation/<int:corporation_id>/<int:year>/view/details/<int:entity_id>/",
        corporation_ledger.corporation_details,
        name="corporation_details",
    ),
    path(
        "corporation/<int:corporation_id>/<int:year>/<int:month>/view/details/<int:entity_id>/",
        corporation_ledger.corporation_details,
        name="corporation_details",
    ),
    path(
        "corporation/<int:corporation_id>/<int:year>/<int:month>/<int:day>/view/details/<int:entity_id>/",
        corporation_ledger.corporation_details,
        name="corporation_details",
    ),
    path(
        "corporation/<int:corporation_id>/division/<int:division_id>/view/details/<int:entity_id>/",
        corporation_ledger.corporation_details,
        name="corporation_details",
    ),
    path(
        "corporation/<int:corporation_id>/division/<int:division_id>/<int:year>/view/details/<int:entity_id>/",
        corporation_ledger.corporation_details,
        name="corporation_details",
    ),
    path(
        "corporation/<int:corporation_id>/division/<int:division_id>/<int:year>/<int:month>/view/details/<int:entity_id>/",
        corporation_ledger.corporation_details,
        name="corporation_details",
    ),
    path(
        "corporation/<int:corporation_id>/division/<int:division_id>/<int:year>/<int:month>/<int:day>/view/details/<int:entity_id>/",
        corporation_ledger.corporation_details,
        name="corporation_details",
    ),
    path(
        "corporation/<int:corporation_id>/view/administration/",
        corporation_ledger.corporation_administration,
        name="corporation_administration",
    ),
    path(
        "corporation/view/overview/",
        corporation_ledger.corporation_overview,
        name="corporation_overview",
    ),
    # -- -- Character Ledger
    path(
        "character/<int:character_id>/",
        character_ledger.character_ledger,
        name="character_ledger",
    ),
    path(
        "character/<int:character_id>/<int:year>/",
        character_ledger.character_ledger,
        name="character_ledger",
    ),
    path(
        "character/<int:character_id>/<int:year>/<int:month>/",
        character_ledger.character_ledger,
        name="character_ledger",
    ),
    path(
        "character/<int:character_id>/<int:year>/<int:month>/<int:day>/",
        character_ledger.character_ledger,
        name="character_ledger",
    ),
    # -- -- Character Details
    path(
        "character/<int:character_id>/<int:year>/view/details/",
        character_ledger.character_details,
        name="character_details",
    ),
    path(
        "character/<int:character_id>/<int:year>/<int:month>/view/details/",
        character_ledger.character_details,
        name="character_details",
    ),
    path(
        "character/<int:character_id>/<int:year>/<int:month>/<int:day>/view/details/",
        character_ledger.character_details,
        name="character_details",
    ),
    # -- -- Character Administration
    path(
        "character/<int:character_id>/view/administration/",
        character_ledger.character_administration,
        name="character_administration",
    ),
    path(
        "character/view/overview/",
        character_ledger.character_overview,
        name="character_overview",
    ),
    # -- -- Planetary
    path(
        "character/view/planetary/",
        planetary.planetary_ledger_index,
        name="planetary_ledger_index",
    ),
    path(
        "character/<int:character_id>/view/planetary/",
        planetary.planetary_ledger,
        name="planetary_ledger",
    ),
    path(
        "character/switch_alarm/",
        planetary.switch_alarm,
        name="switch_alarm",
    ),
    path(
        "character/view/planetary/overview/",
        planetary.planetary_overview,
        name="planetary_overview",
    ),
    # -- API System
    re_path(r"^api/", api.urls),
]



# Generated by Django 4.2.20 on 2025-05-04 21:51

# Django
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0010_fetch_planets_and_update_planetdetails"),
    ]

    operations = []



# Standard Library
from unittest.mock import patch

# Django
from django.test import TestCase, override_settings
from django.utils import timezone

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase

# AA Ledger
from ledger.models.general import EveEntity
from ledger.tests.testdata.esi_stub import esi_client_stub
from ledger.tests.testdata.generate_characteraudit import (
    create_characteraudit_from_evecharacter,
)
from ledger.tests.testdata.generate_walletjournal import create_wallet_journal_entry
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.managers.character_journal_manager"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".esi")
@patch(MODULE_PATH + ".etag_results")
@patch("ledger.models.general.EveEntity")
class TestCharacterJournalManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()
        cls.audit = create_characteraudit_from_evecharacter(1001)

        cls.eve_character_first_party = EveEntity.objects.get(eve_id=1001)
        cls.eve_character_second_party = EveEntity.objects.get(eve_id=1002)

        cls.journal_entry = create_wallet_journal_entry(
            journal_type="character",
            character=cls.audit,
            context_id=1,
            entry_id=10,
            amount=1000,
            balance=2000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2016,
                month=10,
                day=29,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Test Journal",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="player_donation",
        )

    def test_update_wallet_journal(self, mock_eveentity, mock_etag, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        mock_etag.side_effect = lambda ob, token, force_refresh=False: ob.results()

        mock_eveentity.objects.create_bulk_from_esi.return_value = True

        EveEntity.objects.create(
            eve_id=9999, name="Test Character 1", category="character"
        )

        self.audit.update_wallet_journal(force_refresh=False)

        self.assertSetEqual(
            set(self.audit.ledger_character_journal.values_list("entry_id", flat=True)),
            {10, 13, 16},
        )
        obj = self.audit.ledger_character_journal.get(entry_id=10)
        self.assertEqual(obj.amount, 1000)
        self.assertEqual(obj.context_id, 1)
        self.assertEqual(obj.first_party.eve_id, 1001)
        self.assertEqual(obj.second_party.eve_id, 1002)

        obj = self.audit.ledger_character_journal.get(entry_id=13)
        self.assertEqual(obj.amount, 5000)

        obj = self.audit.ledger_character_journal.get(entry_id=16)
        self.assertEqual(obj.amount, 10000)


class TestCharacterJournalManagerAnnotations(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()
        cls.audit = create_characteraudit_from_evecharacter(1001)

        cls.eve_character_first_party = EveEntity.objects.get(eve_id=1001)
        cls.eve_character_second_party = EveEntity.objects.get(eve_id=1002)

        cls.journal_entry = create_wallet_journal_entry(
            journal_type="character",
            character=cls.audit,
            context_id=1,
            entry_id=10,
            amount=1000,
            balance=2000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2016,
                month=10,
                day=29,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Test Journal",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="player_donation",
        )

    def test_annotate_bounty_income(self):
        qs = self.audit.ledger_character_journal.all().annotate_bounty_income()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "bounty_income"),
                "Bounty income annotation should be present",
            )
            self.assertEqual(obj.bounty_income, 0)

    def test_annotate_mission_income(self):
        qs = self.audit.ledger_character_journal.all().annotate_mission_income()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "mission_income"),
                "Mission income annotation should be present",
            )
            self.assertEqual(obj.mission_income, 0)

    def test_annotate_market_income(self):
        qs = self.audit.ledger_character_journal.all().annotate_market_income()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "market_income"),
                "Market income annotation should be present",
            )
            self.assertEqual(obj.market_income, 0)

    def test_annotate_incursion_income(self):
        qs = self.audit.ledger_character_journal.all().annotate_incursion_income()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "incursion_income"),
                "Incursion income annotation should be present",
            )
            self.assertEqual(obj.incursion_income, 0)

    def test_annotate_contract_income(self):
        qs = self.audit.ledger_character_journal.all().annotate_contract_income()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "contract_income"),
                "Contract income annotation should be present",
            )
            self.assertEqual(obj.contract_income, 0)

    def test_annotate_donation_income(self):
        qs = self.audit.ledger_character_journal.all().annotate_donation_income()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "donation_income"),
                "Donation income annotation should be present",
            )
            self.assertEqual(obj.donation_income, 1000.00)

    def test_annotate_insurance_income(self):
        qs = self.audit.ledger_character_journal.all().annotate_insurance_income()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "insurance_income"),
                "Insurance income annotation should be present",
            )
            self.assertEqual(obj.insurance_income, 0)

    def test_annotate_milestone_income(self):
        qs = self.audit.ledger_character_journal.all().annotate_milestone_income()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "milestone_income"),
                "Milestone income annotation should be present",
            )
            self.assertEqual(obj.milestone_income, 0)

    def test_annotate_daily_goal_income(self):
        qs = self.audit.ledger_character_journal.all().annotate_daily_goal_income()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "daily_goal_income"),
                "Daily goal income annotation should be present",
            )
            self.assertEqual(obj.daily_goal_income, 0)

    def test_annotate_miscellaneous_income(self):
        qs = self.audit.ledger_character_journal.all().annotate_miscellaneous()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "miscellaneous"),
                "Miscellaneous income annotation should be present",
            )
            self.assertEqual(obj.miscellaneous, 1000.00)

    def test_annotate_miscellaneous_with_exclude(self):
        qs = (
            self.audit.ledger_character_journal.all().annotate_miscellaneous_with_exclude()
        )
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "miscellaneous"),
                "Miscellaneous with exclude annotation should be present",
            )
            self.assertEqual(obj.miscellaneous, 1000.00)

    def test_annotate_contract_cost(self):
        qs = self.audit.ledger_character_journal.all().annotate_contract_cost()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "contract_cost"),
                "Contract costs annotation should be present",
            )
            self.assertEqual(obj.contract_cost, 0)

    def test_annotate_market_cost(self):
        qs = self.audit.ledger_character_journal.all().annotate_market_cost()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "market_cost"),
                "Market costs annotation should be present",
            )
            self.assertEqual(obj.market_cost, 0)

    def test_annotate_asset_cost(self):
        qs = self.audit.ledger_character_journal.all().annotate_asset_cost()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "asset_cost"),
                "Asset costs annotation should be present",
            )
            self.assertEqual(obj.asset_cost, 0)

    def test_annotate_traveling_cost(self):
        qs = self.audit.ledger_character_journal.all().annotate_traveling_cost()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "traveling_cost"),
                "Traveling costs annotation should be present",
            )
            self.assertEqual(obj.traveling_cost, 0)

    def test_annotate_production_cost(self):
        qs = self.audit.ledger_character_journal.all().annotate_production_cost()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "production_cost"),
                "Production costs annotation should be present",
            )
            self.assertEqual(obj.production_cost, 0)

    def test_annotate_skill_cost(self):
        qs = self.audit.ledger_character_journal.all().annotate_skill_cost()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "skill_cost"),
                "Skill costs annotation should be present",
            )
            self.assertEqual(obj.skill_cost, 0)

    def test_annotate_insurance_cost(self):
        qs = self.audit.ledger_character_journal.all().annotate_insurance_cost()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "insurance_cost"),
                "Insurance costs annotation should be present",
            )
            self.assertEqual(obj.insurance_cost, 0)

    def test_annotate_planetary_cost(self):
        qs = self.audit.ledger_character_journal.all().annotate_planetary_cost()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "planetary_cost"),
                "Planetary costs annotation should be present",
            )
            self.assertEqual(obj.planetary_cost, 0)

    def test_annotate_lp_cost(self):
        qs = self.audit.ledger_character_journal.all().annotate_lp_cost()
        for obj in qs:
            self.assertTrue(
                hasattr(obj, "lp_cost"),
                "LP costs annotation should be present",
            )
            self.assertEqual(obj.lp_cost, 0)


class TestCharacterJournalManagerAggregate(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()
        cls.audit = create_characteraudit_from_evecharacter(1001)

        cls.eve_character_first_party = EveEntity.objects.get(eve_id=1001)
        cls.eve_character_second_party = EveEntity.objects.get(eve_id=1002)

        cls.journal_entry = create_wallet_journal_entry(
            journal_type="character",
            character=cls.audit,
            context_id=1,
            entry_id=10,
            amount=1000,
            balance=2000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2016,
                month=10,
                day=29,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Test Journal",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="player_donation",
        )

    def test_aggregate_bounty(self):
        result = self.audit.ledger_character_journal.all().aggregate_bounty()
        self.assertEqual(result, 0)

    def test_aggregate_costs(self):
        result = self.audit.ledger_character_journal.all().aggregate_costs()
        self.assertEqual(result, 0)

    def test_aggregate_miscellaneous(self):
        result = self.audit.ledger_character_journal.all().aggregate_miscellaneous()
        self.assertEqual(result, 1000.00)

    def test_aggregate_ref_type(self):
        result = self.audit.ledger_character_journal.all().aggregate_ref_type(
            ref_type=["player_donation"], income=True
        )
        self.assertEqual(result, 1000.00)



# Standard Library
from typing import Any

# Django
from django.db import models

# Alliance Auth
from allianceauth.eveonline.providers import ObjectNotFound
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class EveEntityManager(models.Manager):
    def get_or_create_esi(self, *, eve_id: int) -> tuple[Any, bool]:
        """gets or creates entity object with data fetched from ESI"""
        # pylint: disable=import-outside-toplevel
        # AA Ledger
        from ledger.models import EveEntity

        try:
            entity = self.get(eve_id=eve_id)
            return entity, False
        except EveEntity.DoesNotExist:
            return self.update_or_create_esi(eve_id=eve_id)

    def create_bulk_from_esi(self, eve_ids):
        """gets bulk names with ESI"""
        if len(eve_ids) > 0:
            # pylint: disable=import-outside-toplevel
            # AA Ledger
            from ledger.models.general import EveEntity

            chunk_size = 500
            id_chunks = [
                eve_ids[i : i + chunk_size] for i in range(0, len(eve_ids), chunk_size)
            ]
            for chunk in id_chunks:
                response = esi.client.Universe.post_universe_names(ids=chunk).results()
                new_names = []
                logger.debug(
                    "Eve Entity Manager EveName: count in %s count out %s",
                    len(chunk),
                    len(response),
                )
                for entity in response:
                    new_names.append(
                        EveEntity(
                            eve_id=entity["id"],
                            name=entity["name"],
                            category=entity["category"],
                        )
                    )
                EveEntity.objects.bulk_create(new_names, ignore_conflicts=True)
            return True
        return True

    def update_or_create_esi(self, *, eve_id: int) -> tuple[Any, bool]:
        """updates or creates entity object with data fetched from ESI"""
        response = esi.client.Universe.post_universe_names(ids=[eve_id]).results()
        if len(response) != 1:
            raise ObjectNotFound(eve_id, "unknown_type")
        entity_data = response[0]
        return self.update_or_create(
            eve_id=entity_data["id"],
            defaults={
                "name": entity_data["name"],
                "category": entity_data["category"],
            },
        )



"""PvE Views"""

# Standard Library
import json
from decimal import Decimal

# Django
from django.core.cache import cache
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import DecimalField, F, Q, QuerySet, Sum
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.eveonline.models import EveAllianceInfo
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.app_settings import LEDGER_CACHE_ENABLED, LEDGER_CACHE_STALE
from ledger.helpers.core import LedgerCore, LedgerEntity
from ledger.helpers.ref_type import RefTypeManager
from ledger.models.corporationaudit import (
    CorporationAudit,
    CorporationWalletJournalEntry,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class AllianceData(LedgerCore):
    """Class to hold alliance data for the ledger."""

    # pylint: disable=too-many-positional-arguments
    def __init__(
        self,
        request: WSGIRequest,
        alliance: EveAllianceInfo,
        year=None,
        month=None,
        day=None,
    ):
        LedgerCore.__init__(self, year, month, day)
        self.request = request
        self.alliance = alliance
        self.corporations = CorporationAudit.objects.filter(
            corporation__alliance__alliance_id=self.alliance.alliance_id
        ).values_list("corporation__corporation_id", flat=True)
        # Evaluate the existing years for the view
        self.existing_years = (
            CorporationWalletJournalEntry.objects.filter(
                division__corporation__corporation__alliance__alliance_id=self.alliance.alliance_id
            )
            .exclude(date__year__isnull=True)
            .values_list("date__year", flat=True)
            .order_by("-date__year")
            .distinct()
        )
        self.auth_char_ids = self.auth_character_ids

    def setup_ledger(self, entity: LedgerEntity = None):
        """Setup the Ledger Data for the Corporation."""
        alliance_id = self.alliance.alliance_id

        # Base queryset filtered by date and alliance
        base_qs = CorporationWalletJournalEntry.objects.filter(
            self.filter_date,
            division__corporation__corporation__alliance__alliance_id=alliance_id,
        )

        # No entity specified: show all entries for the alliance (preserve existing exclusion)
        if entity is None:
            self.journal = base_qs.exclude(
                Q(ref_type="corporation_account_withdrawal")
                & Q(first_party_id=F("second_party_id"))
            )
            # Prepare auxiliary data used by the view
            self.entities = self.get_all_entities(self.journal)
            return

        # If the entity is the alliance itself and "all" is set, show all alliance entries
        if self.request.GET.get("all", False) and entity.entity_id == alliance_id:
            self.journal = base_qs.exclude(
                Q(ref_type="corporation_account_withdrawal")
                & Q(first_party_id=F("second_party_id"))
            )
            self.entities = self.get_all_entities(self.journal)
            return

        # Regular entity filtering: include any rows where the entity is a first or second party
        entity_ids = entity.get_alts_ids_or_self()
        qs = base_qs.filter(
            Q(first_party_id__in=entity_ids) | Q(second_party_id__in=entity_ids)
        )
        qs = qs.exclude(
            Q(ref_type="corporation_account_withdrawal")
            & Q(first_party_id=F("second_party_id"))
        )

        # If entity represents a corporation or alliance, exclude auth account character IDs
        # that are not part of the current entity to avoid double counting
        if entity.type in ["alliance", "corporation"]:
            exclude_ids = self.auth_char_ids - set(entity_ids)
            qs = qs.exclude(
                Q(first_party_id__in=exclude_ids) | Q(second_party_id__in=exclude_ids)
            )

        self.journal = qs
        self.entities = self.get_all_entities(self.journal)

    def get_all_entities(
        self, journal: QuerySet[CorporationWalletJournalEntry]
    ) -> list[int]:
        """Get all entities in the alliance."""
        entities_ids = set(journal.values_list("second_party_id", flat=True)) | set(
            journal.values_list("first_party_id", flat=True)
        )
        return list(entities_ids)

    # pylint: disable=duplicate-code
    def create_entity_data(
        self,
        entity: LedgerEntity,
    ) -> dict:
        """Create the URL for entity details based on the view type."""
        used_pks = set()
        bounty = Decimal(0)
        ess = Decimal(0)
        miscellaneous = Decimal(0)
        costs = Decimal(0)

        for pk, rows in list(self.entries.items()):
            for row in rows:
                if (
                    row["division__corporation__corporation__corporation_id"]
                    == entity.entity_id
                ):
                    if RefTypeManager.special_cases(
                        row, ids=[], account_char_ids=self.auth_char_ids
                    ):
                        continue
                    bounty += row.get("bounty") or Decimal(0)
                    ess += row.get("ess") or Decimal(0)
                    miscellaneous += row.get("miscellaneous") or Decimal(0)
                    costs += row.get("costs") or Decimal(0)
                    used_pks.add(pk)

        # Remove Used Pks from Entries
        # This is to prevent the entries from being used in the future
        for pk in used_pks:
            self.entries.pop(pk, None)

        misc = miscellaneous
        total = sum([bounty, ess, miscellaneous, costs])

        if total == 0:
            return None

        entity_ledger_info = {
            "entity": entity,
            "ledger": {
                "bounty": bounty,
                "ess": ess,
                "miscellaneous": misc,
                "costs": costs,
                "total": total,
            },
            "type": entity.type,
        }

        return entity_ledger_info

    # pylint: disable=duplicate-code
    def generate_ledger_data(self) -> dict:
        """Generate the ledger data for the alliance."""
        ledger = False
        finished_entities = False

        self.setup_ledger()

        journal = self.journal.values(
            "first_party_id",
            "second_party_id",
            "pk",
            "ref_type",
            "division__corporation__corporation__corporation_id",
        ).annotate(
            bounty=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.BOUNTY_PRIZES),
                output_field=DecimalField(),
            ),
            ess=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.ESS_TRANSFER),
                output_field=DecimalField(),
            ),
            costs=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__lt=0),
                output_field=DecimalField(),
            ),
            miscellaneous=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__gt=0),
                output_field=DecimalField(),
            ),
        )

        # Build up the journal Cache
        ledger_hash = self._get_ledger_journal_hash(self.journal.values_list("pk"))
        ledger_header = self._get_ledger_header(
            ledger_args=f"{self.alliance.alliance_id}",
            year=self.year,
            month=self.month,
            day=self.day,
        )
        cache_header = cache.get(
            ledger_header,
        )
        logger.debug(
            f"Ledger Header: {ledger_header}, Cache Header: {cache_header}, Journal Hash: {ledger_hash}"
        )

        # Check if the journal is up to date
        journal_up_to_date = cache_header == ledger_hash
        ledger_key = self._build_ledger_cache_key(ledger_header)

        # Check if we have newest cached version of the ledger
        if journal_up_to_date and LEDGER_CACHE_ENABLED:
            ledger = cache.get(f"{ledger_key}-data", False)
            finished_entities = cache.get(f"{ledger_key}-finished_entities", False)

        if finished_entities is False or ledger is False:
            ledger = []
            finished_entities = set()

            # Build the entries from the journal
            self.entries = {}
            for row in journal:
                self.entries.setdefault(row["pk"], []).append(row)

            # Build Data for each corporation
            for corporation_id in self.corporations:
                # Create Details URL for the entity
                details_url = self.create_url(
                    viewname="alliance_details",
                    alliance_id=self.alliance.alliance_id,
                    entity_id=corporation_id,
                )

                # Create the LedgerEntity object for the entity
                entity_obj = LedgerEntity(
                    entity_id=corporation_id,
                    details_url=details_url,
                )

                corp_data = self.create_entity_data(
                    entity=entity_obj,
                )

                if corp_data is None:
                    continue

                ledger.append(corp_data)
                finished_entities.add(corporation_id)

        # Create the billboard data
        self.create_rattingbar(list(finished_entities))
        self.create_chord(ledger)

        # Build the context for the ledger view
        context = self._build_context(ledger=ledger)

        cache.set(key=f"{ledger_key}-data", value=ledger, timeout=LEDGER_CACHE_STALE)
        cache.set(
            key=f"{ledger_key}-finished_entities",
            value=finished_entities,
            timeout=LEDGER_CACHE_STALE,
        )
        cache.set(
            key=self._get_ledger_header(
                ledger_args=f"{self.alliance.alliance_id}",
                year=self.year,
                month=self.month,
                day=self.day,
            ),
            value=ledger_hash,
            timeout=None,  # Cache forever until the journal changes
        )
        return context

    def _build_context(self, ledger):
        """Build the context for the ledger view."""
        return {
            "title": f"Alliance Ledger - {self.alliance.alliance_name}",
            "alliance_id": self.alliance.alliance_id,
            "billboard": json.dumps(self.billboard.dict.asdict()),
            "ledger": ledger,
            "years": list(self.existing_years),
            "totals": self._calculate_totals(ledger),
            "view": self.create_view_data(
                viewname="alliance_details",
                alliance_id=self.alliance.alliance_id,
                entity_id=self.alliance.alliance_id,
            ),
        }

    def create_rattingbar(self, entities_ids: list = None):
        """Create the ratting bar for the view."""
        if not entities_ids:
            return

        rattingbar_timeline = self.billboard.create_timeline(self.journal)
        rattingbar = (
            rattingbar_timeline.annotate_bounty_income()
            .annotate_ess_income()
            .annotate_miscellaneous()
        )
        self.billboard.create_or_update_results(rattingbar)
        self._build_xy_chart(title=_("Ratting Bar"))

    def create_chord(self, ledger_data: list[dict]):
        """Create the chord chart for the view."""
        if not ledger_data:
            return

        for entry in ledger_data:
            entity_name = entry["entity"].entity_name
            ledger = entry["ledger"]
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Bounty (Wallet)"),
                value=ledger.get("bounty", 0),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("ESS (Wallet)"),
                value=ledger.get("ess", 0),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Costs (Wallet)"),
                value=abs(ledger.get("costs", 0)),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Miscellaneous (Wallet)"),
                value=abs(ledger.get("miscellaneous", 0)),
            )
        self.billboard.chord_handle_overflow()



# Django
from django.test import RequestFactory, TestCase

# AA Ledger
from ledger.helpers.core import (
    add_info_to_context,
)
from ledger.tests.testdata.generate_corporationaudit import (
    add_corporationaudit_corporation_to_user,
    create_user_from_evecharacter,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.view_helpers.core"


class TestViewHelpers(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001, permissions=["ledger.basic_access"]
        )
        cls.audit = add_corporationaudit_corporation_to_user(
            cls.user, cls.character_ownership.character.character_id
        )

    def test_add_info_to_context(self):
        # given
        request = self.factory.get("/")
        request.user = self.user

        context = {
            "theme": None,
            "issues": [],
        }
        # when
        result = add_info_to_context(request, context)
        # then
        self.assertEqual(result, context)



"""TestView class."""

# Standard Library
from http import HTTPStatus
from unittest.mock import Mock, patch

# Django
from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse

# Alliance Auth (External Libs)
from app_utils.testing import (
    create_user_from_evecharacter,
)

# AA Ledger
from ledger.tests.testdata.generate_characteraudit import (
    add_characteraudit_character_to_user,
    create_user_from_evecharacter_with_access,
)
from ledger.tests.testdata.generate_corporationaudit import (
    add_corporationaudit_corporation_to_user,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse
from ledger.views import index
from ledger.views.alliance import alliance_ledger
from ledger.views.character import character_ledger, planetary
from ledger.views.corporation import corporation_ledger

INDEX_PATH = "ledger.views.index"
CHARLEDGER_PATH = "ledger.views.character.character_ledger"
CORPLEDGER_PATH = "ledger.views.corporation.corporation_ledger"
ALLYLEDGER_PATH = "ledger.views.alliance.alliance_ledger"


@patch(INDEX_PATH + ".messages")
class TestViewIndexAccess(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1002
        )
        cls.superuser, cls.character_ownership = (
            create_user_from_evecharacter_with_access(1001)
        )

    def test_admin(self, mock_messages):
        """Test admin access."""
        # given
        self.superuser.is_superuser = True
        self.superuser.save()

        request = self.factory.get(reverse("ledger:admin"))
        request.user = self.superuser
        # when
        response = index.admin(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Administration")

    def test_admin_no_access(self, mock_messages):
        """Test admin access."""
        # given
        request = self.factory.get(reverse("ledger:admin"))
        request.user = self.user

        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = index.admin(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertTrue(mock_messages.error.called)

    def test_admin_clear_all_etags(self, mock_messages):
        """Test clear all etags."""
        # given
        self.superuser.is_superuser = True
        self.superuser.save()
        request = self.factory.post(
            reverse("ledger:admin"), data={"run_clear_etag": True}
        )
        request.user = self.superuser

        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = index.admin(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        mock_messages.info.assert_called_once_with(request, "Queued Clear All ETags")

    def test_force_refresh(self, mock_messages):
        """Test force refresh."""
        # given
        self.superuser.is_superuser = True
        self.superuser.save()
        request = self.factory.post(
            reverse("ledger:admin"), data={"force_refresh": True}
        )
        request.user = self.superuser

        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = index.admin(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_run_char_updates(self, mock_messages):
        """Test run char updates."""
        # given
        self.superuser.is_superuser = True
        self.superuser.save()
        request = self.factory.post(
            reverse("ledger:admin"), data={"run_char_updates": True}
        )
        request.user = self.superuser

        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = index.admin(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        mock_messages.info.assert_called_once_with(
            request, "Queued Update All Characters"
        )

    def test_run_corp_updates(self, mock_messages):
        """Test Run corp updates."""
        # given
        self.superuser.is_superuser = True
        self.superuser.save()
        request = self.factory.post(
            reverse("ledger:admin"), data={"run_corp_updates": True}
        )
        request.user = self.superuser

        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = index.admin(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        mock_messages.info.assert_called_once_with(
            request, "Queued Update All Corporations"
        )


class TestViewCharacterLedgerAccess(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001
        )
        cls.user2, cls.character_ownership2 = create_user_from_evecharacter_with_access(
            1002
        )
        cls.audit = add_characteraudit_character_to_user(
            cls.user, cls.character_ownership.character.character_id
        )

    def test_view_character_ledger(self):
        """Test view character ledger."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:character_ledger",
                args=[self.character_ownership.character.character_id],
            )
        )
        request.user = self.user
        # when
        response = character_ledger.character_ledger(
            request, self.character_ownership.character.character_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Character Ledger")

    @patch(CHARLEDGER_PATH + ".messages")
    def test_view_character_ledger_without_permission(self, mock_messages):
        """Test view character ledger without permission."""
        # given
        request = self.factory.get(reverse("ledger:character_ledger", args=[1003]))
        request.user = self.user

        # Add session middleware to process the request
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        message_middleware = MessageMiddleware(Mock())
        message_middleware.process_request(request)
        # when
        response = character_ledger.character_ledger(request, 1003)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Character Ledger")
        self.assertTrue(mock_messages.error.called)

    def test_view_character_details(self):
        """Test view character details."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:character_details",
                args=[self.character_ownership.character.character_id, 2025],
            )
        )
        request.user = self.user
        # when
        response = character_ledger.character_details(
            request, self.character_ownership.character.character_id, year=2025
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(
            response,
            "No ratting data found...",
        )

    def test_view_character_details_no_permission(self):
        """Test view character details."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:character_details",
                args=[self.character_ownership.character.character_id, 2025],
            )
        )
        request.user = self.user2

        # when
        response = character_ledger.character_details(
            request, self.character_ownership.character.character_id, year=2025
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Permission Denied")

    def test_view_character_overview(self):
        """Test view character overview."""
        # given
        request = self.factory.get(reverse("ledger:character_overview"))
        request.user = self.user
        # when
        response = character_ledger.character_overview(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Character Overview")

    def test_view_character_administration(self):
        """Test view character administration."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:character_administration",
                args=[self.character_ownership.character.character_id],
            )
        )
        request.user = self.user
        # when
        response = character_ledger.character_administration(
            request, self.character_ownership.character.character_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Administration")

    def test_view_character_administration_withouth_character_id(self):
        """Test view character administration."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:character_administration",
                args=[self.character_ownership.character.character_id],
            )
        )
        request.user = self.user
        # when
        response = character_ledger.character_administration(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Administration")

    @patch(CHARLEDGER_PATH + ".messages")
    def test_view_character_administration_no_permission(self, mock_messages):
        """Test view character administration."""
        # given
        request = self.factory.get(
            reverse("ledger:character_administration", args=[1002])
        )
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = character_ledger.character_administration(request, 1002)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.error.assert_called_once_with(request, "Permission Denied")


class TestViewCorporationLedgerAccess(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
                "ledger.manage_access",
                "ledger.corp_audit_manager",
            ],
        )
        cls.user_no_perm, cls.character_ownership_no_perm = (
            create_user_from_evecharacter(
                1002,
                permissions=[
                    "ledger.basic_access",
                    "ledger.advanced_access",
                    "ledger.manage_access",
                ],
            )
        )
        cls.audit = add_corporationaudit_corporation_to_user(cls.user, 1001)

    def test_view_corporation_ledger_index(self):
        """Test view corporation ledger index."""
        # given
        request = self.factory.get(reverse("ledger:corporation_ledger_index"))
        request.user = self.user
        # when
        response = corporation_ledger.corporation_ledger_index(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)

    def test_view_corporation_ledger(self):
        """Test view corporation ledger."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:corporation_ledger",
                args=[self.character_ownership.character.corporation_id],
            )
        )
        request.user = self.user
        # when
        response = corporation_ledger.corporation_ledger(
            request, self.character_ownership.character.corporation_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Corporation Ledger")

    @patch(CORPLEDGER_PATH + ".messages")
    def test_view_corporation_ledger_not_found(self, mock_messages):
        """Test view corporation ledger not found."""
        # given
        request = self.factory.get(reverse("ledger:corporation_ledger", args=[9999]))
        request.user = self.user

        # Add session middleware to process the request
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        message_middleware = MessageMiddleware(Mock())
        message_middleware.process_request(request)
        # when
        response = corporation_ledger.corporation_ledger(request, 9999)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Corporation Ledger")
        self.assertTrue(mock_messages.info.called)

    @patch(CORPLEDGER_PATH + ".messages")
    def test_view_corporation_ledger_without_permission(self, mock_messages):
        """Test view corporation ledger without permission."""
        # given
        request = self.factory.get(reverse("ledger:corporation_ledger", args=[2001]))
        request.user = self.user_no_perm

        # Add session middleware to process the request
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        message_middleware = MessageMiddleware(Mock())
        message_middleware.process_request(request)
        # when
        response = corporation_ledger.corporation_ledger(request, 2001)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Corporation Ledger")
        self.assertTrue(mock_messages.error.called)

    def test_view_corporation_details(self):
        """Test view corporation details."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:corporation_details",
                args=[self.character_ownership.character.corporation_id, 2025, 1001],
            )
        )
        request.user = self.user
        # when
        response = corporation_ledger.corporation_details(
            request,
            self.character_ownership.character.corporation_id,
            entity_id=1001,
            year=2025,
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "No ratting data found...")

    def test_view_corporation_details_no_permission(self):
        """Test view corporation details."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:corporation_details",
                args=[2001, 2025, 1001],
            )
        )
        request.user = self.user_no_perm

        # when
        response = corporation_ledger.corporation_details(
            request, 2001, entity_id=1001, year=2025
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Permission Denied")

    def test_view_corporation_details_not_found(self):
        """Test view corporation details."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:corporation_details",
                args=[9999, 2025, 9999],
            )
        )
        request.user = self.user_no_perm

        # when
        response = corporation_ledger.corporation_details(
            request, 9999, entity_id=9999, year=2025
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Corporation not found")

    def test_view_corporation_overview(self):
        """Test view corporation overview."""
        # given
        request = self.factory.get(reverse("ledger:corporation_overview"))
        request.user = self.user
        # when
        response = corporation_ledger.corporation_overview(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Corporation Overview")

    def test_view_corporation_administration(self):
        """Test view corporation administration."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:corporation_administration",
                args=[self.character_ownership.character.corporation_id],
            )
        )
        request.user = self.user
        # when
        response = corporation_ledger.corporation_administration(
            request, self.character_ownership.character.corporation_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Administration")

    @patch(CORPLEDGER_PATH + ".messages")
    def test_view_corporation_administration_no_permission(self, mock_messages):
        """Test view corporation administration."""
        # given
        request = self.factory.get(
            reverse("ledger:corporation_administration", args=[2001])
        )
        request.user = self.user_no_perm
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = corporation_ledger.corporation_administration(request, 2001)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.error.assert_called_once_with(request, "Permission Denied")

    @patch(CORPLEDGER_PATH + ".messages")
    def test_view_corporation_administration_corporation_not_found(self, mock_messages):
        """Test view corporation administration."""
        # given
        request = self.factory.get(
            reverse("ledger:corporation_administration", args=[6666])
        )
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = corporation_ledger.corporation_administration(request, 6666)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.info.assert_called_once_with(request, "Corporation not found")


class TestViewAllianceLedgerAccess(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
            ],
        )
        cls.user_no_perm, cls.character_ownership_no_perm = (
            create_user_from_evecharacter(
                1003,
                permissions=[
                    "ledger.basic_access",
                    "ledger.advanced_access",
                ],
            )
        )
        cls.user_admin, cls.character_ownership_admin = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
                "ledger.manage_access",
            ],
        )
        cls.audit = add_corporationaudit_corporation_to_user(cls.user, 1001)
        cls.audit_admin = add_corporationaudit_corporation_to_user(cls.user_admin, 1002)
        cls.user_has_no_alliance, cls.character_ownership_no_alliance = (
            create_user_from_evecharacter(
                1000,
                permissions=[
                    "ledger.basic_access",
                    "ledger.advanced_access",
                ],
            )
        )

    def test_view_alliance_ledger_index(self):
        """Test view alliance ledger index."""
        # given
        request = self.factory.get(reverse("ledger:alliance_ledger_index"))
        request.user = self.user
        # when
        response = alliance_ledger.alliance_ledger_index(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)

    @patch(ALLYLEDGER_PATH + ".messages")
    def test_view_alliance_ledger_index_exception(self, mock_messages):
        """Test view alliance ledger."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:alliance_ledger_index",
            )
        )
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        request.user = self.user_has_no_alliance
        # when
        response = alliance_ledger.alliance_ledger_index(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.error.assert_called_once_with(
            request, "You do not have an alliance."
        )

    def test_view_alliance_ledger(self):
        """Test view alliance ledger."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:alliance_ledger",
                args=[self.character_ownership.character.alliance_id],
            )
        )
        request.user = self.user
        # when
        response = alliance_ledger.alliance_ledger(
            request, self.character_ownership.character.alliance_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Alliance Ledger")

    @patch(ALLYLEDGER_PATH + ".messages")
    def test_view_alliance_ledger_not_found(self, mock_messages):
        """Test view alliance ledger not found."""
        # given
        request = self.factory.get(reverse("ledger:alliance_ledger", args=[9999]))
        request.user = self.user

        # Add session middleware to process the request
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        message_middleware = MessageMiddleware(Mock())
        message_middleware.process_request(request)
        # when
        response = alliance_ledger.alliance_ledger(request, 9999)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Alliance Ledger")
        self.assertTrue(mock_messages.info.called)

    @patch(ALLYLEDGER_PATH + ".messages")
    def test_view_alliance_ledger_without_permission(self, mock_messages):
        """Test view alliance ledger without permission."""
        # given
        request = self.factory.get(reverse("ledger:alliance_ledger", args=[3001]))
        request.user = self.user_no_perm

        # Add session middleware to process the request
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        message_middleware = MessageMiddleware(Mock())
        message_middleware.process_request(request)
        # when
        response = alliance_ledger.alliance_ledger(request, 3001)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Alliance Ledger")
        self.assertTrue(mock_messages.error.called)

    def test_view_alliance_details(self):
        """Test view alliance details."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:alliance_details",
                args=[self.character_ownership.character.alliance_id, 2025, 1001],
            )
        )
        request.user = self.user
        # when
        response = alliance_ledger.alliance_details(
            request,
            self.character_ownership.character.alliance_id,
            entity_id=1001,
            year=2025,
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "No ratting data found...")

    def test_view_alliance_details_no_permission(self):
        """Test view alliance details without permission."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:alliance_details",
                args=[3001, 2025, 2001],
            )
        )
        request.user = self.user_no_perm

        # when
        response = alliance_ledger.alliance_details(
            request, 3001, entity_id=2001, year=2025
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Permission Denied")

    def test_view_alliance_details_not_found(self):
        """Test view alliance details not found."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:alliance_details",
                args=[9999, 2025, 9999],
            )
        )
        request.user = self.user_no_perm

        # when
        response = alliance_ledger.alliance_details(
            request, 9999, entity_id=9999, year=2025
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Alliance not found.")

    def test_view_alliance_overview(self):
        """Test view alliance overview."""
        # given
        request = self.factory.get(reverse("ledger:alliance_overview"))
        request.user = self.user
        # when
        response = alliance_ledger.alliance_overview(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Alliance Overview")

    def test_view_alliance_administration(self):
        """Test view alliance administration."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:alliance_administration",
                args=[self.character_ownership_admin.character.alliance_id],
            )
        )
        request.user = self.user_admin
        # when
        response = alliance_ledger.alliance_administration(
            request, self.character_ownership_admin.character.alliance_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Administration")

    @patch(ALLYLEDGER_PATH + ".messages")
    def test_view_alliance_administration_no_permission(self, mock_messages):
        """Test view alliance administration."""
        # given
        request = self.factory.get(
            reverse("ledger:alliance_administration", args=[3001])
        )
        request.user = self.user_admin
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = alliance_ledger.alliance_administration(request, 3001)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.error.assert_called_once_with(request, "Permission Denied")

    @patch(ALLYLEDGER_PATH + ".messages")
    def test_view_alliance_administration_alliance_not_found(self, mock_messages):
        """Test view alliance administration."""
        # given
        request = self.factory.get(
            reverse("ledger:alliance_administration", args=[6666])
        )
        request.user = self.user_admin
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = alliance_ledger.alliance_administration(request, 6666)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.info.assert_called_once_with(request, "Alliance not found.")


class TestViewPlanetaryLedgerAccess(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
            ],
        )

    def test_view_planetary_ledger_index(self):
        """Test view planetary ledger index."""
        # given
        request = self.factory.get(reverse("ledger:planetary_ledger_index"))
        request.user = self.user
        # when
        response = planetary.planetary_ledger_index(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)

    def test_view_planetary_ledger(self):
        """Test view planetary ledger."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:planetary_ledger",
                args=[self.character_ownership.character.character_id],
            )
        )
        request.user = self.user
        # when
        response = planetary.planetary_ledger(
            request, self.character_ownership.character.character_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Planetary Details")

    def test_view_planetary_ledger_without_character_id(self):
        """Test view character ledger."""
        # given
        request = self.factory.get(
            reverse(
                "ledger:planetary_ledger",
                args=[self.character_ownership.character.character_id],
            )
        )
        request.user = self.user
        # when
        response = planetary.planetary_ledger(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Planetary Details")

    def test_view_planetary_overview(self):
        """Test view planetary overview."""
        # given
        request = self.factory.get(reverse("ledger:planetary_overview"))
        request.user = self.user
        # when
        response = planetary.planetary_overview(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, "Planetary Overview")



"""Hook into Alliance Auth"""

# Django
from django.db.models import Sum
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

# AA Ledger
from ledger import app_settings, urls
from ledger.models.characteraudit import CharacterAudit


class LedgerMenuItem(MenuItemHook):
    """This class ensures only authorized users will see the menu entry"""

    def __init__(self):
        super().__init__(
            f"{app_settings.LEDGER_APP_NAME}",
            "fas fa-book fa-fw",
            "ledger:index",
            navactive=["ledger:"],
        )

    def render(self, request):
        if request.user.has_perm("ledger.basic_access"):
            self.count = (
                CharacterAudit.objects.annotate_total_update_status_user(
                    user=request.user
                )
                .aggregate(total_failed=Sum("num_sections_failed"))
                .get("total_failed", 0)
            )
            return MenuItemHook.render(self, request)
        return ""


@hooks.register("menu_item_hook")
def register_menu():
    """Register the menu item"""

    return LedgerMenuItem()


@hooks.register("url_hook")
def register_urls():
    """Register app urls"""

    return UrlHook(urls, "ledger", r"^ledger/")


@hooks.register("charlink")
def register_charlink_hook():
    return "ledger.thirdparty.charlink_hook"



"""
Constants
"""

# Embed colors
DISCORD_EMBED_COLOR_INFO = 0x5BC0DE
DISCORD_EMBED_COLOR_SUCCESS = 0x5CB85C
DISCORD_EMBED_COLOR_WARNING = 0xF0AD4E
DISCORD_EMBED_COLOR_DANGER = 0xD9534F

# Discord embed color map
DISCORD_EMBED_COLOR_MAP = {
    "info": DISCORD_EMBED_COLOR_INFO,
    "success": DISCORD_EMBED_COLOR_SUCCESS,
    "warning": DISCORD_EMBED_COLOR_WARNING,
    "danger": DISCORD_EMBED_COLOR_DANGER,
}

COMMAND_CENTER = [
    2129,
    2130,
    2131,
    2132,
    2136,
    2137,
    2138,
    2139,
    2140,
    2141,
    2142,
    2143,
    2144,
    2145,
    2146,
    2147,
    2148,
    2149,
    2150,
    2151,
    2152,
    2153,
    2154,
    2155,
    2156,
    2157,
    2158,
    2159,
    2160,
    2254,
    2524,
    2525,
    2533,
    2534,
    2549,
    2550,
    2551,
    2574,
    2576,
    2577,
    2568,
    2581,
    2582,
    2585,
    2586,
]
SPACEPORTS = [2256, 2542, 2543, 2544, 2552, 2555, 2556, 2557]
STORAGE_FACILITY = [2257, 2535, 2536, 2541, 2558, 2560, 2561, 2562]
EXTRACTOR_CONTROL_UNIT = [2848, 3060, 3061, 3062, 3063, 3064, 3067, 3068]
P0_PRODUCTS_SOLID = [2267, 2270, 2272, 2306, 2307]
P0_PRODUCTS_LIQUID_GAS = [2268, 2308, 2309, 2310, 2311]
P0_PRODUCTS_ORGANIC = [2073, 2286, 2287, 2288, 2305]
P0_PRODUCTS = P0_PRODUCTS_SOLID + P0_PRODUCTS_LIQUID_GAS + P0_PRODUCTS_ORGANIC
P1_PRODUCTS = [
    2389,
    2390,
    2392,
    2393,
    2395,
    2396,
    2397,
    2398,
    2399,
    2400,
    2401,
    3645,
    3683,
    3779,
    9828,
]
P2_PRODUCTS = [
    44,
    2312,
    2317,
    2319,
    2321,
    2327,
    2328,
    2329,
    2463,
    3689,
    3691,
    3693,
    3695,
    3697,
    3725,
    3775,
    3828,
    983,
    9832,
    9836,
    9838,
    9840,
    9842,
    15317,
]
P3_PRODUCTS = [
    2344,
    2345,
    2346,
    2348,
    2349,
    2351,
    2352,
    2354,
    2358,
    2360,
    2361,
    2366,
    2367,
    9834,
    9846,
    9848,
    12836,
    17136,
    17392,
    17898,
    28974,
]
P4_PRODUCTS = [2867, 2868, 2869, 2870, 2871, 2872, 2875, 2876]
P5_PRODUCTS = []

NPC_ENTITIES = [
    1000125,  # Concord Bounties (Bounty Prizes, ESS
    1000132,  # Secure Commerce Commission (Market Fees)
    1000413,  # Air Laboratories (Daily Login Rewards, etc.)
]

BOUNTY_PRIZES = ["bounty_prizes"]
ESS_TRANSFER = ["ess_escrow_transfer"]
MISSION_REWARD = ["agent_mission_reward", "agent_mission_time_bonus_reward"]
INCURSION = ["corporate_reward_payout"]

# Cost Ref Types
CONTRACT = [
    "contract_price",
    "contract_collateral",
    "contract_reward_deposited",
    "contract_brokers_fee",
    "contract_sales_tax",
    "contract_price_payment_corp",
    "contract_reward",
    "contract_reward_refund",
    "contract_collateral_refund",
    "contract_deposit_refund",
]
MARKET = [
    "market_escrow",
    "transaction_tax",
    "market_provider_tax",
    "brokers_fee",
    "market_transaction",
]
ASSETS = ["asset_safety_recovery_tax"]
TRAVELING = [
    "structure_gate_jump",
    "jump_clone_activation_fee",
    "jump_clone_installation_fee",
]
PRODUCTION = [
    "industry_job_tax",
    "manufacturing",
    "researching_time_productivity",
    "researching_material_productivity",
    "copying",
    "reprocessing_tax",
    "reaction",
]
SKILL = ["skill_purchase"]
INSURANCE = ["insurance"]
PLANETARY = [
    "planetary_export_tax",
    "planetary_import_tax",
    "planetary_construction",
]
LP = ["lp_store"]
# Trading
DONATION = ["player_donation"]
INSURANCE = ["insurance"]
# MISC
MILESTONE_REWARD = ["milestone_reward_payment"]
DAILY_GOAL_REWARD = ["daily_goal_payouts"]

RENTAL = ["office_rental_fee"]
CORP_WITHDRAW = ["corporation_account_withdrawal"]



# Standard Library
from unittest.mock import PropertyMock, patch

# Django
from django.contrib.auth.models import Permission
from django.test import TestCase
from django.utils import timezone

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# AA Ledger
from ledger.models.corporationaudit import (
    CorporationAudit,
)
from ledger.models.general import EveEntity
from ledger.tests.testdata.generate_characteraudit import (
    add_characteraudit_character_to_user,
    create_user_from_evecharacter_with_access,
)
from ledger.tests.testdata.generate_walletjournal import (
    create_wallet_journal_entry,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.models.corporationaudit"


class TestCharacterWalletJournalModel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()

        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001,
        )
        cls.audit = add_characteraudit_character_to_user(
            cls.user, cls.character_ownership.character.character_id
        )
        cls.eve_character_first_party = EveEntity.objects.get(eve_id=1001)
        cls.eve_character_second_party = EveEntity.objects.get(eve_id=1002)
        cls.journal_entry = create_wallet_journal_entry(
            journal_type="character",
            character=cls.audit,
            entry_id=1,
            amount=1000,
            balance=1000000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2024,
                month=1,
                day=1,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Test",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="test",
        )

    def test_str(self):
        self.assertEqual(
            str(self.journal_entry),
            f"Character Wallet Journal: RefType: test - {self.eve_character_first_party.name} -> {self.eve_character_second_party.name}: 1000 ISK",
        )

    def test_get_visible_should_get_empty_list(self):
        self.assertEqual(
            list(self.journal_entry.get_visible(self.user)), [self.journal_entry]
        )



# Generated by Django 4.2.11 on 2025-01-30 22:59

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0007_characterplanetdetails_facilitys"),
    ]

    operations = [
        migrations.AlterField(
            model_name="characterminingledger",
            name="date",
            field=models.DateTimeField(),
        ),
    ]



"""PvE Views"""

# Standard Library
from http import HTTPStatus

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_POST

# Alliance Auth
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.api.helpers import (
    get_corporation,
    get_manage_corporation,
)
from ledger.helpers.core import add_info_to_context
from ledger.helpers.corporation import CorporationData, LedgerEntity
from ledger.models.corporationaudit import CorporationAudit

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("ledger.advanced_access")
def corporation_ledger_index(request):
    """Corporation Ledger Index View"""
    context = {}
    context = add_info_to_context(request, context)
    return redirect(
        "ledger:corporation_ledger", request.user.profile.main_character.corporation_id
    )


# pylint: disable=too-many-positional-arguments
@login_required
@permission_required("ledger.advanced_access")
def corporation_ledger(
    request: WSGIRequest,
    corporation_id: int,
    division_id: int = None,
    year: int = None,
    month: int = None,
    day: int = None,
):
    """
    Corporation Ledger
    """
    perms, corporation = get_corporation(request, corporation_id)

    context = {
        "title": "Corporation Ledger",
        "corporation_id": corporation_id,
    }

    # pylint: disable=duplicate-code
    if perms is False:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return render(
            request, "ledger/corpledger/corporation_ledger.html", context=context
        )
    # pylint: disable=duplicate-code
    if perms is None:
        msg = _("Corporation not found")
        messages.info(request, msg)
        return render(
            request, "ledger/corpledger/corporation_ledger.html", context=context
        )

    corporation_data = CorporationData(
        request=request,
        corporation=corporation,
        division_id=division_id,
        year=year,
        month=month,
        day=day,
    )

    # Create the Corporation ledger data
    context = corporation_data.generate_ledger_data()
    # Add additional information to the context
    context = add_info_to_context(request, context)

    return render(request, "ledger/corpledger/corporation_ledger.html", context=context)


# pylint: disable=too-many-positional-arguments
@login_required
@permission_required("ledger.advanced_access")
def corporation_details(
    request: WSGIRequest,
    corporation_id: int,
    entity_id: int,
    division_id: int = None,
    year: int = None,
    month: int = None,
    day: int = None,
):
    """
    Corporation Details
    """
    perms, corporation = get_corporation(request, corporation_id)

    context = {
        "title": "Corporation Ledger",
        "corporation_id": corporation_id,
    }

    # pylint: disable=duplicate-code
    if perms is False:
        msg = _("Permission Denied")
        return render(
            request,
            "ledger/partials/information/view_character_content.html",
            {
                "error": msg,
                "corporation_id": corporation_id,
            },
        )
    # pylint: disable=duplicate-code
    if perms is None:
        msg = _("Corporation not found")
        return render(
            request,
            "ledger/partials/information/view_character_content.html",
            {
                "error": msg,
                "corporation_id": corporation_id,
            },
        )

    corporation_data = CorporationData(
        request=request,
        corporation=corporation,
        division_id=division_id,
        year=year,
        month=month,
        day=day,
    )

    # Create the Entity for the ledger
    entity = LedgerEntity(
        entity_id=entity_id,
    )

    amounts = corporation_data._create_corporation_details(entity=entity)
    details = corporation_data._add_average_details(request, amounts, day)

    context = {
        "title": f"Corporation Details - {corporation.corporation_name}",
        "type": "corporation",
        "character": details,
        "information": f"Corporation Details - {corporation_data.get_details_title}",
    }
    context = add_info_to_context(request, context)
    # pylint: disable=duplicate-code
    return render(
        request,
        "ledger/partials/information/view_character_content.html",
        context=context,
    )


@login_required
@permission_required("ledger.advanced_access")
def corporation_overview(request):
    """
    Corporation Overview
    """
    context = {"title": "Corporation Overview"}
    context = add_info_to_context(request, context)
    return render(
        request, "ledger/corpledger/admin/corporation_overview.html", context=context
    )


@login_required
@permission_required("ledger.manage_access")
def corporation_administration(request, corporation_id):
    """
    Corporation Administration
    """
    perm, corporation = get_manage_corporation(request, corporation_id)

    if perm is False:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return redirect("ledger:corporation_ledger_index")
    if perm is None:
        msg = _("Corporation not found")
        messages.info(request, msg)
        return redirect("ledger:corporation_ledger_index")
    # TODO Get Missing Characters from esi-corporations.read_corporation_membership.v1 ?
    corp_characters = CharacterOwnership.objects.filter(
        character__corporation_id=corporation.corporation.corporation_id
    ).order_by("character__character_name")

    context = {
        "corporation_id": corporation_id,
        "title": "Corporation Administration",
        "corporation": corporation,
        "characters": corp_characters,
    }
    context = add_info_to_context(request, context)
    return render(
        request,
        "ledger/corpledger/admin/corporation_administration.html",
        context=context,
    )


@login_required
@permission_required("ledger.manage_access")
@require_POST
def corporation_delete(request, corporation_id):
    """
    Character Delete
    """
    perms = get_manage_corporation(request, corporation_id)[0]

    if perms is False:
        msg = _("Permission Denied")
        return JsonResponse(
            {"success": False, "message": msg}, status=HTTPStatus.FORBIDDEN, safe=False
        )
    if perms is None:
        msg = _("Corporation not found")
        return JsonResponse(
            {"success": False, "message": msg}, status=HTTPStatus.NOT_FOUND, safe=False
        )

    audit = CorporationAudit.objects.get(corporation__corporation_id=corporation_id)
    audit.delete()

    msg = _(f"{audit.corporation.corporation_name} successfully deleted")
    return JsonResponse(
        {"success": True, "message": msg}, status=HTTPStatus.OK, safe=False
    )
